"""Air Purifier Base Class."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from pyvesync.base_devices.vesyncbasedevice import DeviceState, VeSyncBaseDevice

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import AirFryerMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel


logger = logging.getLogger(__name__)


class FryerState(DeviceState):
    """State class for Air Fryer devices.

    Note: This class is a placeholder for future functionality and does not currently
    implement any specific features or attributes.
    """

    __slots__ = ()

    def __init__(
        self,
        device: VeSyncFryer,
        details: ResponseDeviceDetailsModel,
        feature_map: AirFryerMap,
    ) -> None:
        """Initialize FryerState.

        Args:
            device (VeSyncFryer): The device object.
            details (ResponseDeviceDetailsModel): The device details.
            feature_map (AirFryerMap): The feature map for the device.

        """
        super().__init__(device, details, feature_map)
        self.device: VeSyncFryer = device
        self.features: list[str] = feature_map.features


class VeSyncFryer(VeSyncBaseDevice):
    """Base class for VeSync Air Fryer devices."""

    __slots__ = ()

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: AirFryerMap,
    ) -> None:
        """Initialize VeSyncFryer.

        Args:
            details (ResponseDeviceDetailsModel): The device details.
            manager (VeSync): The VeSync manager.
            feature_map (AirFryerMap): The feature map for the device.

        Note:
            This is a bare class as there is only one supported air fryer model.
        """
        super().__init__(details, manager, feature_map)



"""Data models for VeSync outlets."""

from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated, TypeVar

from mashumaro.config import BaseConfig
from mashumaro.mixins.orjson import DataClassORJSONMixin
from mashumaro.types import Alias

from pyvesync.models.base_models import (
    RequestBaseModel,
    ResponseBaseModel,
    ResponseCodeModel,
)
from pyvesync.models.bypass_models import (
    BypassV1Result,
    BypassV2InnerResult,
    RequestBypassV1,
)

T = TypeVar('T', bound='RequestWHOGYearlyEnergy')


@dataclass
class Response7AOutlet(ResponseBaseModel):
    """Response model for 7A outlet."""

    activeTime: int
    energy: float
    deviceStatus: str
    power: float | str
    voltage: float | str


@dataclass
class ResponseEnergyHistory(ResponseCodeModel):
    """Response model for energy history."""

    result: ResponseEnergyResult


@dataclass
class ResponseEnergyResult(BypassV2InnerResult):
    """Response model for energy result."""

    energyInfos: list[EnergyInfo]
    energyConsumptionOfToday: float | None = None
    costPerKWH: float | None = None
    maxEnergy: float | None = None
    totalEnergy: float | None = None


@dataclass
class EnergyInfo:
    """Energy Info list items."""

    timestamp: int
    energyKWH: Annotated[float, Alias('energy')]
    money: float | None = None

    class Config(BaseConfig):
        """Allow deserialization not by alias."""

        allow_deserialization_not_by_alias = True


@dataclass
class Response10ADetails(DataClassORJSONMixin):
    """Response model for Etekcity outlet details."""

    code: int
    msg: str | None
    deviceStatus: str
    connectionStatus: str
    activeTime: int
    power: float
    voltage: float
    energy: float | None = None
    nightLightStatus: str | None = None
    nightLightAutoMode: str | None = None
    nightLightBrightness: int | None = None


@dataclass
class RequestESW03Status(RequestBypassV1):
    """Request model for Etekcity 10A outlet status."""

    status: str


@dataclass
class ResponseESW03Details(BypassV1Result):
    """Response model for Etekcity 10A outlet details."""

    activeTime: int
    deviceStatus: str
    connectionStatus: str
    power: float | None = None
    voltage: float | None = None
    energy: float | None = None


@dataclass
class ResponseOldEnergy(ResponseCodeModel):
    """Response model for old energy history."""

    energyConsumptionOfToday: float
    costPerKWH: float
    maxEnergy: float
    totalEnergy: float
    data: list[float]


@dataclass
class Response15ADetails(ResponseCodeModel):
    """Response for 15A Outlets."""

    result: Response15AOutletResult


@dataclass
class Response15AOutletResult(ResponseBaseModel):
    """Response model for 15A outlet."""

    deviceStatus: str
    connectionStatus: str
    activeTime: int
    power: float
    voltage: float
    energy: float | None = None
    nightLightStatus: str | None = None
    nightLightAutoMode: str | None = None
    nightLightBrightness: int | None = None


@dataclass
class Request15ADetails(RequestBypassV1):
    """Request data model for 15A outlet Details."""


@dataclass
class RequestOutdoorStatus(RequestBypassV1):
    """Request model for outlet status."""

    status: str
    switchNo: str


@dataclass
class RequestEnergyHistory(RequestBaseModel):
    """Request model for energy history."""

    acceptLanguage: str
    appVersion: str
    accountID: str
    method: str
    phoneBrand: str
    phoneOS: str
    timeZone: str
    token: str
    traceId: str
    userCountryCode: str
    debugMode: bool
    homeTimeZone: str
    uuid: str


@dataclass
class Request15AStatus(RequestBypassV1):
    """Request data model for 15A outlet.

    Inherits from RequestBypassV1.
    """

    status: str


@dataclass
class Request15ANightlight(RequestBypassV1):
    """Nightlight request data model for 15A Outlets.

    Inherits from RequestBypassV1.
    """

    mode: str


@dataclass
class ResponseOutdoorDetails(ResponseCodeModel):
    """Response model for outdoor outlet."""

    result: ResponseOutdoorOutletResult


@dataclass
class ResponseOutdoorOutletResult(ResponseBaseModel):
    """Response model for outdoor outlet."""

    deviceStatus: str
    connectionStatus: str
    activeTime: int
    power: float
    voltage: float
    energy: float
    subDevices: list[ResponseOutdoorSubDevices]


@dataclass
class ResponseOutdoorSubDevices(ResponseBaseModel):
    """Response model for outdoor energy."""

    subDeviceNo: int
    defaultName: str
    subDeviceName: str
    subDeviceStatus: str


@dataclass
class ResponseBSDGO1Details(ResponseCodeModel):
    """Response model for BSDGO1 outlet."""

    result: ResponseBSDGO1OutletResult


@dataclass
class ResponseWHOGResult(BypassV2InnerResult):
    """Response model for WHOG outlet."""

    enabled: bool
    voltage: float | None = None
    power: float | None = None
    energy: float | None = None
    current: float | None = None
    highestVoltage: float | None = None
    voltagePtStatus: bool | None = None


@dataclass
class ResponseBSDGO1OutletResult(BypassV2InnerResult):
    """Response model for BSDGO1 outlet."""

    powerSwitch_1: int
    realTimeVoltage: float | None = None
    realTimePower: float | None = None
    electricalEnergy: float | None = None
    voltageUpperThreshold: float | None = None
    protectionStatus: str | None = None
    currentUpperThreshold: float | None = None


@dataclass
class Timer7AItem(ResponseBaseModel):
    """Timer item for 7A outlet."""

    timerID: str
    counterTimer: int
    action: str
    timerStatus: str


@dataclass
class ResultESW10Details(ResponseBaseModel):
    """Response model for ESW10 outlet."""

    enabled: bool


@dataclass
class RequestWHOGYearlyEnergy(RequestBypassV1):
    """Request model for WHOG yearly energy history."""

    subDeviceNo: int = 0

    def __post_serialize__(self: RequestWHOGYearlyEnergy, d: dict) -> dict:
        """Remove unneeded keys after serialization."""
        d.pop('uuid', None)
        return d



"""Data models for VeSync Purifiers.

These models inherit from `ResponseBaseModel` and `RequestBaseModel` from the
`base_models` module.

The `InnerPurifierBaseResult` class is used as a base class for the inner purifier
result models for all models and the mashumaro discriminator determines the correct
subclass when deserializing.
"""

from __future__ import annotations

from dataclasses import dataclass

from mashumaro.config import BaseConfig
from mashumaro.types import Discriminator

from pyvesync.models.base_models import (
    RequestBaseModel,
    ResponseBaseModel,
)
from pyvesync.models.bypass_models import (
    BypassV1Result,
    BypassV2InnerResult,
    RequestBypassV1,
)


@dataclass
class InnerPurifierBaseResult(BypassV2InnerResult):
    """Base class for inner purifier results model."""

    class Config(BaseConfig):  # type: ignore[override]
        """Configure the results model to use subclass discriminator."""

        discriminator = Discriminator(include_subtypes=True)


@dataclass
class PurifierVitalDetailsResult(InnerPurifierBaseResult):
    """Vital 100S/200S and Everest Purifier Result Model."""

    powerSwitch: int
    filterLifePercent: int
    workMode: str
    manualSpeedLevel: int
    fanSpeedLevel: int
    AQLevel: int
    PM25: int
    screenState: int
    childLockSwitch: int
    screenSwitch: int
    lightDetectionSwitch: int
    environmentLightState: int
    scheduleCount: int
    timerRemain: int
    efficientModeTimeRemain: int
    errorCode: int
    autoPreference: V2AutoPreferences | None = None
    fanRotateAngle: int | None = None
    filterOpenState: int | None = None
    PM1: int | None = None
    PM10: int | None = None
    AQPercent: int | None = None


@dataclass
class V2AutoPreferences:
    """Vital 100S/200S Auto Preferences."""

    autoPreferenceType: str
    roomSize: int


@dataclass
class PurifierSproutResult(InnerPurifierBaseResult):
    """Sprout Purifier Result Model."""

    powerSwitch: int
    workMode: str
    manualSpeedLevel: int | None
    fanSpeedLevel: int | None
    PM1: int | None
    PM25: int | None
    PM10: int | None
    screenState: int
    childLockSwitch: int
    screenSwitch: int
    scheduleCount: int
    timerRemain: int
    humidity: int | None
    AQI: int | None
    AQLevel: int | None
    temperature: int | None
    VOC: int | None
    CO2: int | None
    errorCode: int
    nightlight: PurifierNightlight | None = None
    autoPreference: V2AutoPreferences | None = None


@dataclass
class PurifierNightlight(ResponseBaseModel):
    """Purifier Nightlight Response Dict."""

    nightLightSwitch: bool
    brightness: int
    colorTemperature: int


@dataclass
class PurifierCoreDetailsResult(InnerPurifierBaseResult):
    """Purifier inner Result Dict."""

    enabled: bool
    filter_life: int
    mode: str
    level: int
    device_error_code: int
    levelNew: int | None = None
    air_quality: int | None = None
    display: bool | None = None
    child_lock: bool | None = None
    configuration: PurifierCoreDetailsConfig | None = None
    extension: dict | None = None
    air_quality_value: int | None = None
    night_light: str | None = None
    fan_rotate: str | None = None


# Purifier Timer Models


@dataclass
class PurifierModifyTimerResult(InnerPurifierBaseResult):
    """Purifier inner Add Timer Result Dict."""

    id: int


@dataclass
class PurifierGetTimerResult(InnerPurifierBaseResult):
    """Purifier inner Timer Result Dict."""

    timers: list[ResponsePurifierTimerItems] | None


@dataclass
class ResponsePurifierTimerItems(ResponseBaseModel):
    """Purifier Timer Items Response Dict."""

    id: int
    remain: int
    total: int
    action: str


@dataclass
class PurifierV2TimerPayloadData(RequestBaseModel):
    """Purifier Timer Payload Data Request Dict."""

    enabled: bool
    startAct: list[PurifierV2TimerActionItems]
    tmgEvt: PurifierV2EventTiming | None = None
    type: int = 0
    subDeviceNo: int = 0
    repeat: int = 0


@dataclass
class PurifierV2TimerActionItems(RequestBaseModel):
    """Purifier Timer Action Items Request Dict."""

    type: str
    act: int
    num: int = 0


@dataclass
class PurifierV2EventTiming(RequestBaseModel):
    """Purifier Event Timing Request Dict."""

    clkSec: int


# Internal Purifier Details Models


@dataclass
class PurifierCoreDetailsConfig(ResponseBaseModel):
    """Config dict in Core purifier details response."""

    display: bool
    display_forever: bool
    auto_preference: None | PurifierCoreAutoConfig = None


@dataclass
class PurifierCoreAutoConfig(ResponseBaseModel):
    """Auto configuration Core dict in purifier details response."""

    type: str
    room_size: int


@dataclass
class PurifierDetailsExtension(ResponseBaseModel):
    """Extension dict in purifier details response for Core 200/300/400."""

    schedule_count: int
    timer_remain: int


# LV - PUR131S Purifier Models


@dataclass
class RequestPurifier131(RequestBypassV1):
    """Purifier 131 Request Dict."""

    status: str | None = None


@dataclass
class RequestPurifier131Mode(RequestBypassV1):
    """Purifier 131 Request Dict."""

    mode: str


@dataclass
class RequestPurifier131Level(RequestBypassV1):
    """Purifier 131 Request Dict."""

    level: int


@dataclass
class Purifier131Result(BypassV1Result):
    """Purifier 131 Details Response Dict."""

    screenStatus: str
    activeTime: int
    levelNew: int
    level: int | None
    mode: str
    airQuality: str
    deviceName: str
    deviceStatus: str
    connectionStatus: str
    filterLife: Purifier131Filter | None = None


@dataclass
class Purifier131Filter(ResponseBaseModel):
    """Filter details model for LV PUR131."""

    change: bool
    useHour: int
    percent: int


@dataclass
class PurifierRH131Result(BypassV2InnerResult):
    """Purifier RH131 Details Response Dict."""

    enabled: bool
    filter_life: int
    mode: str
    level: int
    air_quality: int
    display: bool
    child_lock: bool
    buzzer: bool
    configuration: dict
    extension: PurifierRH131DetailsExtension | None = None


@dataclass
class PurifierRH131DetailsExtension(ResponseBaseModel):
    """Extension dict in RH131 purifier details response."""

    schedule_count: int
    timer_remain: int



"""Test scripts for Etekcity Outdoor Outlet."""
from typing import Any, Dict, Union
from copy import deepcopy
import orjson
from pyvesync.devices.vesyncoutlet import VeSyncOutdoorPlug
from pyvesync.utils.helpers import Helpers as Helpers
import call_json
import call_json_outlets
from base_test_cases import TestBase
from defaults import TestDefaults

DEVICE_TYPE = 'ESO15-TB'

DEV_LIST_DETAIL: Dict[str, Union[str, int, float]
                      ] = call_json.DeviceList.device_list_item(DEVICE_TYPE, 0)

DEV_LIST_DETAIL_2: Dict[str, Any] = call_json.DeviceList.device_list_item(DEVICE_TYPE, 1)

CORRECT_OUTDOOR_LIST: Dict[str, Any] = deepcopy(call_json.DeviceList.list_response_base)
CORRECT_OUTDOOR_LIST['result']['list'].extend([DEV_LIST_DETAIL, DEV_LIST_DETAIL_2])
CORRECT_OUTDOOR_RESP: tuple = (CORRECT_OUTDOOR_LIST, 200)


ENERGY_HISTORY: tuple = call_json_outlets.ENERGY_HISTORY

CORRECT_OUTDOOR_DETAILS = call_json_outlets.DETAILS_RESPONSES[DEVICE_TYPE]

BAD_OUTDOOR_LIST: tuple = call_json.DETAILS_BADCODE

DEFAULTS = TestDefaults


class TestVesyncOutdoorPlug(TestBase):
    """Test class for outdoor outlet."""

    def test_outdoor_conf(self):
        """Tests outdoor outlet is instantiated properly."""
        resp_dict, status = CORRECT_OUTDOOR_RESP
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        self.run_in_loop(self.manager.get_devices)
        outlets = self.manager.outlets
        assert len(outlets) == 2
        outdoor_outlet = outlets[0]
        assert isinstance(outdoor_outlet, VeSyncOutdoorPlug)
        assert outdoor_outlet.device_type == DEVICE_TYPE
        assert outdoor_outlet.uuid == DEFAULTS.uuid(DEVICE_TYPE)

    def test_outdoor_details(self):
        """Tests retrieving outdoor outlet details."""
        resp_dict, status = CORRECT_OUTDOOR_DETAILS
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        outdoor_outlet = VeSyncOutdoorPlug(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(outdoor_outlet.get_details)
        dev_details = outdoor_outlet.details
        assert outdoor_outlet.device_status == 'on'
        assert isinstance(outdoor_outlet, VeSyncOutdoorPlug)
        assert dev_details['active_time'] == 1

    def test_outdoor_details_fail(self, caplog):
        """Test outdoor outlet get_details response."""
        resp_dict, status = BAD_OUTDOOR_LIST
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        outdoor_outlet = VeSyncOutdoorPlug(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(outdoor_outlet.get_details)
        assert len(caplog.records) == 1
        assert 'details' in caplog.text

    def test_outdoor_outlet_onoff(self):
        """Test Outdoor Outlet Device On/Off Methods."""
        self.mock_api.return_value = (orjson.dumps({'code': 0}), 200)
        outdoor_outlet = VeSyncOutdoorPlug(DEV_LIST_DETAIL, self.manager)
        head = Helpers.req_legacy_headers(self.manager)
        body = Helpers.req_body(self.manager, 'devicestatus')

        body['status'] = 'on'
        body['uuid'] = outdoor_outlet.uuid
        body['switchNo'] = outdoor_outlet.sub_device_no
        on = self.run_in_loop(outdoor_outlet.turn_on)
        self.mock_api.assert_called_with(
            '/outdoorsocket15a/v1/device/devicestatus', 'put',
            headers=head, json_object=body
        )
        assert on
        off = self.run_in_loop(outdoor_outlet.turn_off)
        body['status'] = 'off'
        self.mock_api.assert_called_with(
            '/outdoorsocket15a/v1/device/devicestatus', 'put',
            headers=head, json_object=body
        )
        assert off

    def test_outdoor_outlet_onoff_fail(self):
        """Test outdoor outlet On/Off Fail with Code>0."""
        self.mock_api.return_value = (orjson.dumps({'code': 1}), 400)
        outdoor_outlet = VeSyncOutdoorPlug(DEV_LIST_DETAIL, self.manager)
        assert not self.run_in_loop(outdoor_outlet.turn_on)
        assert not self.run_in_loop(outdoor_outlet.turn_off)

    def test_outdoor_outlet_weekly(self):
        """Test outdoor outlet get_weekly_energy."""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        outdoor_outlet = VeSyncOutdoorPlug(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(outdoor_outlet.get_weekly_energy)
        body = Helpers.req_body(self.manager, 'energy_week')
        body['uuid'] = outdoor_outlet.uuid
        self.mock_api.assert_called_with(
            '/outdoorsocket15a/v1/device/energyweek',
            'post',
            headers=Helpers.req_legacy_headers(self.manager),
            json_object=body,
        )
        energy_dict = outdoor_outlet.energy['week']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]
        assert outdoor_outlet.weekly_energy_total == 1

    def test_outdoor_outlet_monthly(self):
        """Test outdoor outlet get_monthly_energy."""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        outdoor_outlet = VeSyncOutdoorPlug(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(outdoor_outlet.get_monthly_energy)
        body = Helpers.req_body(self.manager, 'energy_month')
        body['uuid'] = outdoor_outlet.uuid
        self.mock_api.assert_called_with(
            '/outdoorsocket15a/v1/device/energymonth',
            'post',
            headers=Helpers.req_legacy_headers(self.manager),
            json_object=body,
        )
        energy_dict = outdoor_outlet.energy['month']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]
        assert outdoor_outlet.monthly_energy_total == 1

    def test_outdoor_outlet_yearly(self):
        """Test outdoor outlet get_yearly_energy."""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        outdoor_outlet = VeSyncOutdoorPlug(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(outdoor_outlet.get_yearly_energy)
        body = Helpers.req_body(self.manager, 'energy_year')
        body['uuid'] = outdoor_outlet.uuid
        self.mock_api.assert_called_with(
            '/outdoorsocket15a/v1/device/energyyear',
            'post',
            headers=Helpers.req_legacy_headers(self.manager),
            json_object=body,
        )
        energy_dict = outdoor_outlet.energy['year']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]
        assert outdoor_outlet.yearly_energy_total == 1

    def test_history_fail(self):
        """Test outdoor outlet energy failure."""
        bad_history = orjson.dumps({'code': 1})
        self.mock_api.return_value = (bad_history, 200)
        outdoor_outlet = VeSyncOutdoorPlug(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(outdoor_outlet.update_energy)
        assert len(self.caplog.records) == 1
        assert 'weekly' in self.caplog.text
        self.caplog.clear()
        self.run_in_loop(outdoor_outlet.get_monthly_energy)
        assert len(self.caplog.records) == 1
        assert 'monthly' in self.caplog.text
        self.caplog.clear()
        self.run_in_loop(outdoor_outlet.get_yearly_energy)
        assert len(self.caplog.records) == 1
        assert 'yearly' in self.caplog.text



"""library_logger.py.

This module provides a custom logger for libraries. It automatically
captures the caller's module and class name in the log records, and it
provides helper methods for logging API calls and other specially formatted
messages.

Usage:
    from library_logger import LibraryLogger

    # Use `LibraryLogger.log_api_call` to log API calls.
    LibraryLogger.log_api_call(logger, response)

    # Use `LibraryLogger.log_api_exception` to log API exceptions.
    LibraryLogger.log_api_exception(logger, request_dict, exception)
"""

from __future__ import annotations

import logging
import os
import re
import sys
from collections.abc import Mapping
from dataclasses import fields, is_dataclass
from pathlib import Path
from typing import TYPE_CHECKING

import orjson
from mashumaro.exceptions import InvalidFieldValue, MissingField, UnserializableField
from multidict import CIMultiDictProxy

from pyvesync.utils.errors import ErrorTypes, ResponseInfo

if TYPE_CHECKING:
    from aiohttp import ClientResponse
    from aiohttp.client_exceptions import ClientResponseError

    from pyvesync.base_devices.vesyncbasedevice import VeSyncBaseDevice


class LibraryLogger:
    """Library Logging Interface.

    LibraryLogger provides a logging interface that automatically adds context
    (module and class name) and offers helper methods for structured logging,
    such as API call logs.

    Attributes:
        debug (bool): Class attribute to enable or disable debug logging,
            prints API requests & responses that return an error.
        shouldredact (bool): Class attribute whether to redact
            sensitive information from logs.
        verbose (bool): Class attribute to print all request & response content.

    Examples:
        Logging an API call:
        >>> LibraryLogger.log_api_call(logger, response)
            2025-02-01 12:34:56 - DEBUG - my_library - ========API CALL========
            API CALL to endpoint: /api/v1/resource
            Method: POST
            Request Headers:
            {
                REQUEST HEADERS
            }
            Request Body:
                {
                    "key": "value"
                }
            Response Body:
                {
                    RESPONSE BODY
                }

        Logging an API exception:
        >>> LibraryLogger.log_api_exception(logger, request_dict, exception)
            2025-02-01 12:34:56 - DEBUG - pyvesync -
            Error in API CALL to endpoint: /api/v1/resource
            Response Status: 500
            Exception: <Exception>
            Method: POST
            Request Headers:
                {
                    REQUEST HEADERS
                }
            Request Body:
                {
                    "key": "value"
                }
    """

    shouldredact = True
    """Class attribute to determine if sensitive information should be redacted."""
    verbose = False
    """Class attribute to print all request & response content."""
    debug_enabled = False
    """Class attribute to cache status of debug logging to avoid expensive calls."""

    @classmethod
    def check_debug(cls) -> bool:
        """Check if debug logging is enabled."""
        cls.debug_enabled = logging.getLogger('pyvesync').isEnabledFor(logging.DEBUG)
        return cls.debug_enabled

    @classmethod
    def redactor(cls, stringvalue: str) -> str:
        """Redact sensitive strings from debug output.

        This method searches for specific sensitive keys in the input string and replaces
        their values with '##_REDACTED_##'. The keys that are redacted include:

        - token
        - authorizeCode
        - password
        - email
        - tk
        - accountId
        - authKey
        - uuid
        - cid

        Args:
            stringvalue (str): The input string potentially containing
                sensitive information.

        Returns:
            str: The redacted string with sensitive information replaced
                by '##_REDACTED_##'.
        """
        if cls.shouldredact:
            stringvalue = re.sub(
                (
                    r'(?i)'
                    r'((?<=token":\s")|'
                    r'(?<=password":\s")|'
                    r'(?<=email":\s")|'
                    r'(?<=tk":\s")|'
                    r'(?<=accountId":\s")|'
                    r'(?<=accountID":\s")|'
                    r'(?<=authKey":\s")|'
                    r'(?<=uuid":\s")|'
                    r'(?<=cid":\s")|'
                    r'(?<=token\s)|'
                    r'(?<=authorizeCode\s)|'
                    r'(?<=account_id\s))'
                    r'[^"\s]+'
                ),
                '##_REDACTED_##',
                stringvalue,
            )
        return stringvalue

    @staticmethod
    def is_json(data: str | bytes | None) -> bool:
        """Check if the data is JSON formatted."""
        if data is None:
            return False
        if isinstance(data, str):
            data = data.encode('utf-8')
        try:
            orjson.loads(data)
        except orjson.JSONDecodeError:
            return False
        return True

    @staticmethod
    def try_json_loads(data: str | bytes | None) -> dict | None:
        """Try to load JSON data.

        Gracefully handle errors and return None if loading fails.

        Args:
            data (str | bytes | None): JSON data to load.

        Returns:
            dict | None: Parsed JSON data or None if loading fails.
        """
        if data is None:
            return None
        try:
            return orjson.loads(data)
        except (orjson.JSONDecodeError, TypeError):
            return None

    @classmethod
    def api_printer(cls, api: Mapping | bytes | str | None) -> str | None:
        """Print the API dictionary in a readable format."""
        if api is None or len(api) == 0:
            return None
        try:
            if isinstance(api, bytes):
                api_dict = orjson.loads(api)
            elif isinstance(api, str):
                api_dict = orjson.loads(api.encode('utf-8'))
            elif isinstance(api, (dict, CIMultiDictProxy)):
                api_dict = dict(api)
            else:
                return None
            dump = orjson.dumps(
                dict(api_dict), option=orjson.OPT_INDENT_2 | orjson.OPT_NON_STR_KEYS
            )
            return cls.redactor(dump.decode('utf-8'))

        except (orjson.JSONDecodeError, orjson.JSONEncodeError):
            if isinstance(api, bytes):
                return api.decode('utf-8')
            return str(api)

    @staticmethod
    def set_log_level(level: str | int = logging.WARNING) -> None:
        """Set log level for logger to INFO."""
        # Test default log levels if int or str
        logging.getLogger().setLevel(level)

    @staticmethod
    def configure_file_logging(
        log_file: str | Path,
        *,
        level: str | int = logging.INFO,
        stdout: bool = True,
        propagate: bool = False,
    ) -> None:
        """Configure logging for pyvesync library.

        Args:
            log_file (str | Path | None): The name of the file to log to. If None,
                logs will only be printed to the console.
            level (str | int): The log level to set the logger to, can be
                in form of enum `logging.DEBUG` or string `DEBUG`.
            stdout (bool): If True, log messages will also be printed to stdout.
            propagate (bool): If True, log messages will propagate to the root logger.

        Note:
            This method configures the pyvesync base logger and sets
            propagation of log messages to the root logger based on the
            `propagate` parameter.
        """
        logger = logging.getLogger('pyvesync')
        logger.setLevel(level)
        logger.propagate = propagate

        log_path = Path(log_file)
        log_path.parent.mkdir(parents=True, exist_ok=True)
        file_handler = logging.FileHandler(log_path.resolve())
        stream_handler = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(
            fmt='%(asctime)s - %(levelname)s - %(name)s - %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
        )
        file_handler.setFormatter(formatter)
        logger.addHandler(file_handler)
        if stdout:
            stream_handler.setFormatter(formatter)
            logger.addHandler(stream_handler)

    @classmethod
    def error_device_response_code(
        cls,
        logger: logging.Logger,
        device: VeSyncBaseDevice,
        method_name: str,
        code: int | str,
        msg: str,
    ) -> None:
        """Log an error response from a device API call.

        Use this log message to indicate that a device API call
        returned a response with an error.

        Args:
            logger (logging.Logger): module logger
            device (VeSyncBaseDevice): device instance
            method_name (str): device method name
            code (int | str): error code returned
            msg (str): error message
        """
        logger.error(
            '%s (%s) - %s %s returned error code %s in response with msg: %s',
            device.product_type,
            device.device_type,
            device.device_name,
            method_name,
            str(code),
            msg,
        )

    @classmethod
    def error_mashumaro_response(
        cls,
        logger: logging.Logger,
        method_name: str,
        resp_dict: dict,
        exc: MissingField | InvalidFieldValue | UnserializableField,
        device: VeSyncBaseDevice | None = None,
    ) -> None:
        """Log an error parsing API response.

        Use this log message to indicate that the API response
        is not in the expected format.

        Args:
            logger (logging.Logger): module logger
            method_name (str): device name
            resp_dict (dict): API response dictionary
            exc (MissingField | InvalidFieldValue | UnserializableField): mashumaro
                exception caught
            device (VeSyncBaseDevice | None): device if a device method was called
        """
        if not cls.debug_enabled:
            if device is not None:
                logger.error(
                    'Error parsing %s %s %s response with data model %s',
                    device.product_type,
                    device.device_name,
                    method_name,
                    exc.holder_class_name,
                )
            else:
                logger.error(
                    'Error parsing %s response with data model %s',
                    method_name,
                    exc.holder_class_name,
                )
        msg = f'Error parsing {method_name} response data model {exc.holder_class_name}. '
        if isinstance(exc, MissingField):
            msg += f'Missing field: {exc.field_name} of type {exc.field_type_name}'
        elif isinstance(exc, InvalidFieldValue):
            msg += f'Invalid field value: {exc.field_name} of type {exc.field_type_name}'
        elif isinstance(exc, UnserializableField):
            msg += f'Unserializable field: {exc.field_name} of type {exc.field_type_name}'
        msg += (
            '\n\n Please report this issue tohttps://github.com/webdjoe/pyvesync/issues'
        )
        logger.warning(msg)
        if not cls.debug_enabled:
            return
        msg = ''
        if is_dataclass(exc.holder_class):
            holder = exc.holder_class
            field_tuple = fields(holder)
            resp_f = sorted(resp_dict.keys())
            field_f = sorted([f.name for f in field_tuple])
            dif = set(field_f) - set(resp_f)
            if dif:
                msg += '\n\n-------------------------------------'
                msg += '\n Expected Fields:'
                msg += f'\n({", ".join(field_f)})'
                msg += '\n Response Fields:'
                msg += f'\n({", ".join(resp_f)})'
                msg += '\n Missing Fields:'
                msg += f'\n({", ".join(dif)})'

        msg += '\n\n Full Response:'
        msg += f'\n{orjson.dumps(resp_dict, option=orjson.OPT_INDENT_2).decode("utf-8")}'
        msg += '\n\n---------------------------------'
        msg += '\n\n Exception:'
        msg += f'\n{exc.__traceback__}'
        logger.debug(msg)

    @classmethod
    def error_device_response_content(
        cls,
        logger: logging.Logger,
        device: VeSyncBaseDevice,
        method: str,
        msg: str | None = None,
    ) -> None:
        """Log an error parsing API response.

        Use this log message to indicate that the API response
        is not in the expected format.

        Args:
            logger (logging.Logger): module logger
            device (VeSyncBaseDevice): device instance
            method (str): method that caused the error
            msg (str | None, optional): optional description of error
        """
        logger.error(
            '%s (%s - %s) returned an invalid response format in %s method: %s',
            device.device_name,
            device.product_type,
            device.device_type,
            method,
            msg if msg is not None else '',
        )

    @classmethod
    def log_device_return_code(
        cls,
        logger: logging.Logger,
        method: str,
        device_name: str,
        device_type: str,
        error_info: ResponseInfo,
    ) -> None:
        """Log return code from device API call.

        When API responds with JSON, if the code key is not 0,
        it indicates an error has occurred.

        Args:
            logger (logging.Logger): module logger instance
            method (str): method that caused the error
            device_name (str): device name
            device_type (str): device type
            error_info (ResponseInfo): response info dataclass
        """
        try:
            code_str = str(error_info.code)
        except (TypeError, ValueError):
            code_str = 'UNKNOWN'
        if error_info.code == 0:
            logger.debug(
                '%s for %s API from %s returned code: %s, message: %s',
                device_name,
                device_type,
                method,
                code_str,
                error_info.message,
            )
        elif error_info.error_type == ErrorTypes.DEVICE_OFFLINE:
            logger.debug(
                '%s for %s API from %s returned DEVICE OFFLINE code: %s, message: %s',
                device_name,
                device_type,
                method,
                code_str,
                error_info.message,
            )
        else:
            logger.warning(
                (
                    '%s for %s API from %s returned error code: %s,'
                    ' error type: %s, message: %s'
                ),
                device_name,
                device_type,
                method,
                code_str,
                error_info.error_type,
                error_info.message,
            )

    @staticmethod
    def _resolve_base_caller_of_async_call() -> str:  # noqa: C901
        """Resolve the frame that invoked VeSync.async_call_api."""
        try:
            pkg_root = 'pyvesync'

            # Start from the caller of log_api_call
            f = sys._getframe(1)  # noqa: SLF001  # pylint: disable=protected-access

            # Find the frame for async_call_api in pyvesync.vesync
            async_frame = None
            while f:
                mod_name = f.f_globals.get('__name__', '')
                func_name = f.f_code.co_name
                if func_name == 'async_call_api' and (
                    mod_name == f'{pkg_root}.vesync' or mod_name.endswith('.vesync')
                ):
                    async_frame = f
                    break
                f = f.f_back  # type: ignore[assignment]

            # The "base caller" is the immediate caller of async_call_api
            target = async_frame.f_back if async_frame else None
            if not target:
                return 'unknown'

            # Skip any frames inside our logging util if present
            while target and target.f_globals.get('__name__', '').startswith(
                f'{pkg_root}.utils.logs'
            ):
                target = target.f_back
            if not target:
                return 'unknown'

            # Build a readable identifier
            func_name = target.f_code.co_name
            mod_name = target.f_globals.get('__name__', '')

            # Prefer class context if available
            cls_name = None
            if 'self' in target.f_locals and target.f_locals['self'] is not None:
                cls = type(target.f_locals['self'])
                cls_name = getattr(cls, '__name__', None)
                mod_name = getattr(cls, '__module__', mod_name)
            elif 'cls' in target.f_locals and target.f_locals['cls'] is not None:
                cls = target.f_locals['cls']
                cls_name = getattr(cls, '__name__', None)
                mod_name = getattr(cls, '__module__', mod_name)

            # Normalize module for compactness: strip root package
            human_mod = mod_name
            if human_mod.startswith(f'{pkg_root}.'):
                human_mod = human_mod[len(pkg_root) + 1 :]  # noqa: E203, RUF100

            if cls_name:
                return f'{cls_name}.{func_name} [{human_mod}]'

        except Exception:  # noqa: BLE001
            return 'unknown'
        return f'{human_mod}.{func_name}'

    @classmethod
    def log_api_call(
        cls,
        logger: logging.Logger,
        response: ClientResponse,
        response_body: bytes | None = None,
        request_headers: dict | None = None,
        request_body: str | dict | None = None,
    ) -> None:
        """Log API calls in debug mode.

        Logs an API call with a specific format that includes the endpoint,
        JSON-formatted headers, request body (if any) and response body.

        Args:
            logger (logging.Logger): The logger instance to use.
            response (aiohttp.ClientResponse): Requests response object from the API call.
            response_body (bytes, optional): The response body to log.
            request_headers (dict, optional): The request headers to log.
            request_body (dict | str, optional): The request body to log.

        Notes:
            This is a method used for the logging of API calls when the debug
            flag is enabled. The method logs the endpoint, method, request headers,
            request body (if any), response headers, and response body (if any).
        """
        if cls.debug_enabled is False:
            return
        # Build the log message parts.

        # Emit a dedicated debug line for the base caller
        parts = ['==================API CALL==================']
        try:
            base_caller = cls._resolve_base_caller_of_async_call()
            # Keep message simple to avoid breaking existing structured logs
            parts.append(f'Caller: {base_caller}')
        except Exception:  # noqa: BLE001,S110
            pass

        parts.append(f'API CALL to endpoint: {response.url.path}')
        parts.append(f'Host: {response.url.host}')
        parts.append(f'Full URL: {response.url}')
        parts.append(f'Response Status: {response.status}')
        parts.append(f'Method: {response.method}')
        parts.append('---------------Request-----------------')

        if request_headers:
            parts.append(
                f'Request Headers: {os.linesep} {cls.api_printer(request_headers)}'
            )
        if request_body is not None:
            request_body = cls.api_printer(request_body)
            parts.append(f'Request Body: {os.linesep} {request_body}')

        parts.append('---------------Response-----------------')
        response_headers = cls.api_printer(response.headers)
        if response_headers:
            parts.append(f'Response Headers: {os.linesep} {response_headers}')

        response_dict = cls.try_json_loads(response_body)
        if response_dict is not None:
            response_str = cls.api_printer(response_dict)
            parts.append(f'Response Body: {os.linesep} {response_str}')
        elif isinstance(response_body, bytes) and len(response_body) > 0:
            response_str = response_body.decode('utf-8')
            parts.append(f'Error parsing response body: {os.linesep} {response_str}')

        full_message = os.linesep.join(parts)
        logger.debug(full_message)

    @classmethod
    def log_api_status_error(
        cls,
        logger: logging.Logger,
        *,
        response: ClientResponse,
    ) -> None:
        """Log API response with non-200 status codes.

        Args:
            logger (logging.Logger): The logger instance to use.
            response (aiohttp.ClientResponse): KW only, dictionary
                containing the request information.
        """
        # Build the log message parts.
        msg = (
            f'Status Code {response.status} error in {response.method}'
            f' API CALL to endpoint: {response.url.path}'
        )
        logger.error(msg)

    @classmethod
    def log_api_exception(
        cls,
        logger: logging.Logger,
        *,
        exception: ClientResponseError,
        request_body: dict | None,
    ) -> None:
        """Log asyncio response exceptions.

        Logs an API call with a specific format that includes the endpoint,
        JSON-formatted headers, request body (if any) and response body.

        Args:
            logger (logging.Logger): The logger instance to use.
            exception (ClientResponseError): KW only, The request body to log.
            request_body (dict | None): KW only, The request body.
        """
        # Build the log message parts.
        parts = [
            f'asyncio error in API CALL to endpoint: {exception.request_info.url.path}'
        ]
        parts.append(f'Exception Raised: {exception}')
        req_headers = cls.api_printer(exception.request_info.headers)
        if req_headers is not None:
            parts.append(f'Request Headers: {os.linesep} {req_headers}')

        req_body = cls.api_printer(request_body)
        if req_body is not None:
            parts.append(f'Request Body: {os.linesep} {req_body}')

        full_message = os.linesep.join(parts)
        logger.error(full_message)



"""Air Purifier Base Class."""

from __future__ import annotations

import logging
from abc import abstractmethod
from typing import TYPE_CHECKING

from typing_extensions import deprecated

from pyvesync.base_devices.vesyncbasedevice import DeviceState, VeSyncBaseToggleDevice
from pyvesync.const import (
    AirQualityLevel,
    DeviceStatus,
    NightlightModes,
    PurifierFeatures,
    PurifierModes,
)

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import PurifierMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel


logger = logging.getLogger(__name__)


class PurifierState(DeviceState):
    """Base state class for Purifiers.

    Attributes:
        active_time (int): Active time of device, defaults to None.
        connection_status (str): Connection status of device.
        device (VeSyncOutlet): Device object.
        device_status (str): Device status.
        features (dict): Features of device.
        last_update_ts (int): Last update timestamp of device, defaults to None.
        mode (str): Current mode of the purifier.
        fan_level (int): Current fan level of the purifier.
        fan_set_level (int): Set fan level of the purifier.
        filter_life (int): Filter life percentage of the purifier.
        auto_preference_type (str): Auto preference type of the purifier.
        auto_room_size (int): Auto room size of the purifier.
        air_quality_level (AirQualityLevel): Air quality level of the purifier.
        child_lock (bool): Child lock status of the purifier.
        display_status (str): Display status of the purifier.
        display_set_status (str): Display set status of the purifier.
        display_forever (bool): Display forever status of the purifier.
        timer (Timer): Timer object for the purifier.
        pm25 (int): PM2.5 value of the purifier.
        pm1 (int): PM1 value of the purifier.
        pm10 (int): PM10 value of the purifier.
        aq_percent (int): Air quality percentage of the purifier.
        light_detection_switch (str): Light detection switch status of the purifier.
        light_detection_status (str): Light detection status of the purifier.
        nightlight_status (str): Nightlight status of the purifier.
        fan_rotate_angle (int): Fan rotate angle of the purifier.
        temperature (int): Temperature value of the purifier.
        humidity (int): Humidity value of the purifier.
        voc (int): VOC value of the purifier.
        co2 (int): CO2 value of the purifier.
        nightlight_brightness (int): Nightlight brightness level of the purifier.

    Note:
        Not all attributes are supported by all models.
    """

    __slots__ = (
        '_air_quality_level',
        'aq_percent',
        'auto_preference_type',
        'auto_room_size',
        'child_lock',
        'co2',
        'display_forever',
        'display_set_status',
        'display_status',
        'fan_level',
        'fan_rotate_angle',
        'fan_set_level',
        'filter_life',
        'filter_open_state',
        'humidity',
        'light_detection_status',
        'light_detection_switch',
        'mode',
        'nightlight_brightness',
        'nightlight_status',
        'pm1',
        'pm10',
        'pm25',
        'temperature',
        'voc',
    )

    def __init__(
        self,
        device: VeSyncPurifier,
        details: ResponseDeviceDetailsModel,
        feature_map: PurifierMap,
    ) -> None:
        """Initialize Purifier State."""
        super().__init__(device, details, feature_map)
        self.mode: str = PurifierModes.UNKNOWN
        self.fan_level: int | None = None
        self.fan_set_level: int | None = None
        self.filter_life: int | None = None
        self.auto_preference_type: str | None = None
        self.auto_room_size: int | None = None
        self._air_quality_level: AirQualityLevel | None = None
        self.child_lock: bool | None = None
        self.filter_open_state: bool = False
        self.display_status: str | None = None
        self.display_set_status: str | None = None
        self.display_forever: bool = False
        self.humidity: int | None = None
        self.temperature: int | None = None
        # Attributes not supported by all purifiers
        self.pm25: int | None = None
        self.pm1: int | None = None
        self.pm10: int | None = None
        self.aq_percent: int | None = None
        self.voc: int | None = None
        self.co2: int | None = None
        self.light_detection_switch: str | None = None
        self.light_detection_status: str | None = None
        self.nightlight_brightness: int | None = None
        self.nightlight_status: str | None = None
        self.fan_rotate_angle: int | None = None

    @property
    def air_quality_level(self) -> int | None:
        """Return air quality level in integer from 1-4.

        Returns None if unknown.
        """
        if self._air_quality_level is None:
            return None
        return int(self._air_quality_level)

    @air_quality_level.setter
    def air_quality_level(self, value: int | None) -> None:
        """Set air quality level."""
        if isinstance(value, int):
            self._air_quality_level = AirQualityLevel.from_int(value)
        elif value is None:
            self._air_quality_level = None

    def set_air_quality_level(self, value: int | str | None) -> None:
        """Set air quality level."""
        if isinstance(value, str):
            self._air_quality_level = AirQualityLevel.from_string(value)
        elif isinstance(value, int):
            self._air_quality_level = AirQualityLevel.from_int(value)
        elif value is None:
            self._air_quality_level = None

    @property
    def air_quality_string(self) -> str | None:
        """Return air quality level as string."""
        if self._air_quality_level is None:
            return None
        return str(self._air_quality_level)

    @property
    @deprecated('Use state.air_quality_level instead.')
    def air_quality(self) -> int | str | None:
        """Return air quality level."""
        return self.air_quality_level

    @property
    @deprecated('Use light_detection_switch instead.')
    def light_detection(self) -> bool:
        """Return light detection status."""
        return self.light_detection_switch == DeviceStatus.ON

    @property
    @deprecated('Use state.pm25 instead.')
    def air_quality_value(self) -> int | None:
        """Return air quality value."""
        return self.pm25

    @property
    @deprecated('Use PurifierState.fan_level instead.')
    def speed(self) -> int | None:
        """Return fan speed."""
        return self.fan_level

    @property
    @deprecated('Use PurifierState.nightlight_status instead.')
    def night_light(self) -> str | None:
        """Return night light status."""
        return self.nightlight_status

    @property
    @deprecated('Use display_status instead.')
    def display_state(self) -> str | None:
        """Return display status."""
        return self.display_status

    @property
    @deprecated('Use display_set_status instead.')
    def display_switch(self) -> str | None:
        """Return display switch status."""
        return self.display_set_status


class VeSyncPurifier(VeSyncBaseToggleDevice):
    """Base device for vesync air purifiers.

    Args:
        details (ResponseDeviceDetailsModel): Device details from API.
        manager (VeSync): VeSync manager instance.
        feature_map (PurifierMap): Feature map for the device.

    Attributes:
        state (PurifierState): State of the device.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
        modes (list[str]): List of modes supported by the device.
        fan_levels (list[int]): List of fan levels supported by the device.
        nightlight_modes (list[str]): List of nightlight modes supported by the device.
        auto_preferences (list[str]): List of auto preferences supported by the device.
    """

    __slots__ = ('auto_preferences', 'fan_levels', 'modes', 'nightlight_modes')

    def __init__(
        self,
        details: ResponseDeviceDetailsModel,
        manager: VeSync,
        feature_map: PurifierMap,
    ) -> None:
        """Initialize VeSync Purifier Base Class."""
        super().__init__(details, manager, feature_map)
        self.features: list[str] = feature_map.features
        self.state: PurifierState = PurifierState(self, details, feature_map)
        self.modes: list[str] = feature_map.modes
        self.fan_levels: list[int] = feature_map.fan_levels
        self.nightlight_modes: list[str] = feature_map.nightlight_modes
        self.auto_preferences: list[str] = feature_map.auto_preferences

    @property
    def supports_air_quality(self) -> bool:
        """Return True if device supports air quality."""
        return PurifierFeatures.AIR_QUALITY in self.features

    @property
    def supports_fan_rotate(self) -> bool:
        """Return True if device supports fan rotation."""
        return PurifierFeatures.VENT_ANGLE in self.features

    @property
    def supports_nightlight(self) -> bool:
        """Return True if device supports nightlight."""
        return PurifierFeatures.NIGHTLIGHT in self.features

    @property
    def supports_light_detection(self) -> bool:
        """Returns True if device supports light detection."""
        return PurifierFeatures.LIGHT_DETECT in self.features

    async def toggle_display(self, mode: bool) -> bool:
        """Set Display Mode."""
        del mode
        raise NotImplementedError

    async def turn_on_display(self) -> bool:
        """Turn on Display."""
        return await self.toggle_display(True)

    async def turn_off_display(self) -> bool:
        """Turn off Display."""
        return await self.toggle_display(False)

    async def set_nightlight_mode(self, mode: str) -> bool:
        """Set Nightlight Mode.

        Modes are defined in the `device.nightlight_modes` attribute.

        Args:
            mode (str): Nightlight mode to set.

        Returns:
            bool: True if successful, False otherwise.
        """
        del mode
        if not self.supports_nightlight:
            logger.error('Nightlight not supported for this device.')
        else:
            logger.error('Nightlight not configured for this device.')
        return False

    async def set_nightlight_dim(self) -> bool:
        """Set Nightlight Dim."""
        return await self.set_nightlight_mode(NightlightModes.DIM)

    async def turn_on_nightlight(self) -> bool:
        """Turn on Nightlight."""
        return await self.set_nightlight_mode(NightlightModes.ON)

    async def turn_off_nightlight(self) -> bool:
        """Turn off Nightlight."""
        return await self.set_nightlight_mode(NightlightModes.OFF)

    async def toggle_child_lock(self, toggle: bool | None = None) -> bool:
        """Toggle Child Lock (Display Lock).

        Args:
            toggle (bool | None): Toggle child lock. If None, toggle state.
        """
        del toggle
        logger.debug('Child lock not configured for this device.')
        return False

    async def turn_on_child_lock(self) -> bool:
        """Set child lock (display lock) to ON."""
        return await self.toggle_child_lock(True)

    async def turn_off_child_lock(self) -> bool:
        """Set child lock (display lock) to OFF."""
        return await self.toggle_child_lock(False)

    @abstractmethod
    async def set_mode(self, mode: str) -> bool:
        """Set Purifier Mode.

        Allowed modes are found in the `device.modes` attribute.

        Args:
            mode (str): Mode to set.

        Returns:
            bool: True if successful, False otherwise.
        """

    @abstractmethod
    async def set_fan_speed(self, speed: int | None = None) -> bool:
        """Set Purifier Fan Speed.

        Args:
            speed (int | None): Fan speed to set. If None, use default speed.

        Returns:
            bool: True if successful, False otherwise.
        """

    async def set_auto_mode(self) -> bool:
        """Set Purifier to Auto Mode."""
        if PurifierModes.AUTO in self.modes:
            return await self.set_mode(PurifierModes.AUTO)
        logger.error('Auto mode not supported for this device.')
        return False

    async def set_sleep_mode(self) -> bool:
        """Set Purifier to Sleep Mode."""
        if PurifierModes.SLEEP in self.modes:
            return await self.set_mode(PurifierModes.SLEEP)
        logger.error('Sleep mode not supported for this device.')
        return False

    async def set_manual_mode(self) -> bool:
        """Set Purifier to Manual Mode."""
        if PurifierModes.MANUAL in self.modes:
            return await self.set_mode(PurifierModes.MANUAL)
        logger.error('Manual mode not supported for this device.')
        return False

    async def set_turbo_mode(self) -> bool:
        """Set Purifier to Turbo Mode."""
        if PurifierModes.TURBO in self.modes:
            return await self.set_mode(PurifierModes.TURBO)
        logger.error('Turbo mode not supported for this device.')
        return False

    async def set_pet_mode(self) -> bool:
        """Set Purifier to Pet Mode."""
        if PurifierModes.PET in self.modes:
            return await self.set_mode(PurifierModes.PET)
        logger.error('Pet mode not supported for this device.')
        return False

    async def set_auto_preference(self, preference: str, room_size: int = 800) -> bool:
        """Set auto preference.

        Args:
            preference (str): Auto preference to set, available preference is
                found in `self.auto_preferences`.
            room_size (int): Room size to set, defaults to 800ft2.

        Returns:
            bool: True if successful, False otherwise.
        """
        del preference, room_size
        logger.debug('Auto preference not configured for this device.')
        return False

    async def toggle_light_detection(self, toggle: bool | None = None) -> bool:
        """Set Light Detection Mode.

        Args:
            toggle (bool | None): Toggle light detection. If None, toggle state.

        Returns:
            bool: True if successful, False otherwise.
        """
        del toggle
        if not self.supports_light_detection:
            logger.debug('Light detection not supported for this device.')
        else:
            logger.debug('Light detection not configured for this device.')
        return False

    async def turn_on_light_detection(self) -> bool:
        """Turn on Light Detection."""
        return await self.toggle_light_detection(True)

    async def turn_off_light_detection(self) -> bool:
        """Turn off Light Detection."""
        return await self.toggle_light_detection(False)

    async def reset_filter(self) -> bool:
        """Reset filter life."""
        logger.debug('Filter life reset not configured for this device.')
        return False

    @deprecated('Use set_auto_mode instead.')
    async def auto_mode(self) -> bool:
        """Set Purifier to Auto Mode."""
        return await self.set_auto_mode()

    @deprecated('Use set_sleep_mode instead.')
    async def sleep_mode(self) -> bool:
        """Set Purifier to Sleep Mode."""
        return await self.set_sleep_mode()

    @deprecated('Use set_manual_mode instead.')
    async def manual_mode(self) -> bool:
        """Set Purifier to Manual Mode."""
        return await self.set_manual_mode()

    @deprecated('Use set_turbo_mode instead.')
    async def turbo_mode(self) -> bool:
        """Set Purifier to Turbo Mode."""
        return await self.set_turbo_mode()

    @deprecated('Use set_pet_mode instead.')
    async def pet_mode(self) -> bool:
        """Set Purifier to Pet Mode."""
        return await self.set_pet_mode()

    @deprecated('Use set_nightlight_mode instead.')
    async def nightlight_mode(self, mode: str) -> bool:
        """Set Nightlight Mode."""
        return await self.set_nightlight_mode(mode)

    @deprecated('Use `set_fan_speed()` instead.')
    async def change_fan_speed(self, speed: int | None = None) -> bool:
        """Deprecated - Set fan speed."""
        return await self.set_fan_speed(speed)

    @deprecated('Use `set_mode(mode: str)` instead.')
    async def change_mode(self, mode: str) -> bool:
        """Deprecated - Set purifier mode."""
        return await self.set_mode(mode)

    @deprecated('Use `toggle_child_lock()` instead.')
    async def set_child_lock(self, toggle: bool) -> bool:
        """Set child lock (display lock).

        This has been deprecated in favor of `toggle_child_lock()`.
        """
        return await self.toggle_child_lock(toggle)

    @deprecated('Use `turn_on_child_lock()` instead.')
    async def child_lock_on(self) -> bool:
        """Turn on child lock (display lock).

        This has been deprecated, use `turn_on_child_lock()` instead.
        """
        return await self.toggle_child_lock(True)

    @deprecated('Use turn_on_light_detection() instead.')
    async def set_light_detection_on(self) -> bool:
        """Turn on light detection feature."""
        return await self.toggle_light_detection(True)

    @deprecated('Use turn_off_light_detection() instead.')
    async def set_light_detection_off(self) -> bool:
        """Turn off light detection feature."""
        return await self.toggle_light_detection(False)



"""Base classes for all VeSync switches."""

from __future__ import annotations

import logging
from typing import TYPE_CHECKING

from typing_extensions import deprecated

from pyvesync.base_devices.vesyncbasedevice import DeviceState, VeSyncBaseToggleDevice
from pyvesync.const import SwitchFeatures
from pyvesync.utils.colors import Color

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.device_map import SwitchMap
    from pyvesync.models.vesync_models import ResponseDeviceDetailsModel
    from pyvesync.utils.colors import HSV, RGB

logger = logging.getLogger(__name__)


class SwitchState(DeviceState):
    """VeSync Switch State Base.

    Args:
        device (VeSyncSwitch): The switch device.
        details (ResponseDeviceDetailsModel): The switch device details.
        feature_map (SwitchMap): The switch feature map.

    Attributes:
        _exclude_serialization (list[str]): List of attributes to exclude from
            serialization.
        active_time (int): Active time of device, defaults to None.
        connection_status (str): Connection status of device.
        device (VeSyncBaseDevice): Device object.
        device_status (str): Device status.
        features (dict): Features of device.
        last_update_ts (int): Last update timestamp of device, defaults to None.
        backlight_color (Color): The backlight color of the switch.
        brightness (int): The brightness level of the switch.
        backlight_status (str): The status of the backlight.
        indicator_status (str): The status of the indicator light.
    """

    __slots__ = (
        '_backlight_color',
        'backlight_status',
        'brightness',
        'indicator_status',
    )

    def __init__(
        self,
        device: VeSyncSwitch,
        details: ResponseDeviceDetailsModel,
        feature_map: SwitchMap,
    ) -> None:
        """Initialize VeSync Switch State."""
        super().__init__(device, details, feature_map)
        self.device: VeSyncSwitch = device
        self._backlight_color: Color | None = None
        self.brightness: int | None = None
        self.active_time: int | None = 0
        self.backlight_status: str | None = None
        self.indicator_status: str | None = None

    @property
    def backlight_rgb(self) -> RGB | None:
        """Get backlight RGB color."""
        if not self.device.supports_backlight_color:
            logger.warning('Backlight color not supported.')
        if isinstance(self._backlight_color, Color):
            return self._backlight_color.rgb
        return None

    @property
    def backlight_hsv(self) -> HSV | None:
        """Get backlight HSV color."""
        if not self.device.supports_backlight_color:
            logger.warning('Backlight color not supported.')
        if isinstance(self._backlight_color, Color):
            return self._backlight_color.hsv
        return None

    @property
    def backlight_color(self) -> Color | None:
        """Get backlight color."""
        if isinstance(self._backlight_color, Color):
            return self._backlight_color
        logger.warning('Backlight color not supported.')
        return None

    @backlight_color.setter
    def backlight_color(self, color: Color | None) -> None:
        """Set backlight color."""
        if not self.device.supports_backlight_color:
            logger.warning('Backlight color not supported.')
            return
        self._backlight_color = color


class VeSyncSwitch(VeSyncBaseToggleDevice):
    """Etekcity Switch Base Class.

    Abstract Base Class for Etekcity Switch Devices, inheriting from
    pyvesync.base_devices.vesyncbasedevice.VeSyncBaseDevice. Should not be
    instantiated directly, subclassed by VeSyncWallSwitch and VeSyncDimmerSwitch.

    Attributes:
        state (SwitchState): Switch state object.
        last_response (ResponseInfo): Last response from API call.
        manager (VeSync): Manager object for API calls.
        device_name (str): Name of device.
        device_image (str): URL for device image.
        cid (str): Device ID.
        connection_type (str): Connection type of device.
        device_type (str): Type of device.
        type (str): Type of device.
        uuid (str): UUID of device, not always present.
        config_module (str): Configuration module of device.
        mac_id (str): MAC ID of device.
        current_firm_version (str): Current firmware version of device.
        device_region (str): Region of device. (US, EU, etc.)
        pid (str): Product ID of device, pulled by some devices on update.
        sub_device_no (int): Sub-device number of device.
        product_type (str): Product type of device.
        features (dict): Features of device.
    """

    __slots__ = ()

    def __init__(
        self, details: ResponseDeviceDetailsModel, manager: VeSync, feature_map: SwitchMap
    ) -> None:
        """Initialize Switch Base Class."""
        super().__init__(details, manager, feature_map)
        self.state: SwitchState = SwitchState(self, details, feature_map)

    @property
    @deprecated('Use `supports_dimmable` property instead.')
    def is_dimmable(self) -> bool:
        """Return True if switch is dimmable."""
        return bool(SwitchFeatures.DIMMABLE in self.features)

    @property
    def supports_backlight_color(self) -> bool:
        """Return True if switch supports backlight."""
        return bool(SwitchFeatures.BACKLIGHT_RGB in self.features)

    @property
    def supports_indicator_light(self) -> bool:
        """Return True if switch supports indicator."""
        return bool(SwitchFeatures.INDICATOR_LIGHT in self.features)

    @property
    def supports_backlight(self) -> bool:
        """Return True if switch supports backlight."""
        return bool(SwitchFeatures.BACKLIGHT in self.features)

    @property
    def supports_dimmable(self) -> bool:
        """Return True if switch is dimmable."""
        return bool(SwitchFeatures.DIMMABLE in self.features)

    async def toggle_indicator_light(self, toggle: bool | None = None) -> bool:
        """Toggle indicator light on or off.

        Args:
            toggle (bool): True to turn on, False to turn off. If None, toggles the state

        Returns:
            bool: True if successful, False otherwise.
        """
        del toggle
        if self.supports_indicator_light:
            logger.debug('toggle_indicator_light not configured for %s', self.device_name)
        else:
            logger.debug('toggle_indicator_light not supported for %s', self.device_name)
        return False

    async def turn_on_indicator_light(self) -> bool:
        """Turn on indicator light if supported."""
        return await self.toggle_indicator_light(True)

    async def turn_off_indicator_light(self) -> bool:
        """Turn off indicator light if supported."""
        return await self.toggle_indicator_light(False)

    async def set_backlight_status(
        self,
        status: bool,
        red: int | None = None,
        green: int | None = None,
        blue: int | None = None,
    ) -> bool:
        """Set the backlight status and optionally its color if supported by the device.

        Args:
            status (bool): Backlight status (True for ON, False for OFF).
            red (int | None): RGB green value (0-255), defaults to None.
            green (int | None): RGB green value (0-255), defaults to None.
            blue (int | None): RGB blue value (0-255), defaults to None.

        Returns:
            bool: True if successful, False otherwise.

        """
        del status, red, green, blue
        if self.supports_backlight:
            logger.debug('set_backlight_status not configured for %s', self.device_name)
        else:
            logger.debug('set_backlight_status not supported for %s', self.device_name)
        return False

    async def turn_on_rgb_backlight(self) -> bool:
        """Turn on backlight if supported."""
        return await self.set_backlight_status(True)

    async def turn_off_rgb_backlight(self) -> bool:
        """Turn off backlight if supported."""
        return await self.set_backlight_status(False)

    @deprecated('Use `turn_on_rgb_backlight()` instead.')
    async def turn_rgb_backlight_on(self) -> bool:
        """Turn on RGB backlight if supported."""
        return await self.set_backlight_status(True)

    @deprecated('Use `turn_off_rgb_backlight()` instead.')
    async def turn_rgb_backlight_off(self) -> bool:
        """Turn off RGB backlight if supported."""
        return await self.set_backlight_status(False)

    async def set_backlight_color(self, red: int, green: int, blue: int) -> bool:
        """Set the color of the backlight using RGB.

        Args:
            red (int): Red value (0-255).
            green (int): Green value (0-255).
            blue (int): Blue value (0-255).

        Returns:
            bool: True if successful, False otherwise.
        """
        return await self.set_backlight_status(True, red=red, green=green, blue=blue)

    async def set_brightness(self, brightness: int) -> bool:
        """Set the brightness of the switch if supported.

        Args:
            brightness (int): Brightness value (0-100).

        Returns:
            bool: True if successful, False otherwise.
        """
        del brightness
        if self.supports_dimmable:
            logger.debug('set_brightness not configured for %s', self.device_name)
        else:
            logger.debug('set_brightness not supported for %s', self.device_name)
        return False

    @deprecated('Use `turn_on_indicator_light` instead.')
    async def turn_indicator_light_on(self) -> bool:
        """Deprecated - use turn_on_indicator_light."""
        return await self.toggle_indicator_light(True)

    @deprecated('Use `turn_off_indicator_light` instead.')
    async def turn_indicator_light_off(self) -> bool:
        """Deprecated - use turn_off_indicator_light."""
        return await self.toggle_indicator_light(False)



"""VeSync Authentication Module."""

from __future__ import annotations

import asyncio
import logging
from pathlib import Path
from typing import TYPE_CHECKING

import orjson
from mashumaro.exceptions import MissingField, UnserializableDataError

from pyvesync.const import DEFAULT_REGION, NON_EU_COUNTRY_CODES
from pyvesync.models.vesync_models import (
    RequestGetTokenModel,
    RequestLoginTokenModel,
    RespGetTokenResultModel,
    RespLoginTokenResultModel,
    ResponseLoginModel,
)
from pyvesync.utils.errors import (
    ErrorCodes,
    ErrorTypes,
    VeSyncAPIResponseError,
    VeSyncLoginError,
    VeSyncServerError,
)

if TYPE_CHECKING:
    from pyvesync.vesync import VeSync

logger = logging.getLogger(__name__)


class VeSyncAuth:
    """VeSync Authentication Manager.

    Handles login, token management, and persistent storage of authentication
    credentials for VeSync API access.

    Args:
        manager: VeSync manager instance for API calls
        username: VeSync account username (email)
        password: VeSync account password
        country_code: Country code in ISO 3166 Alpha-2 format

    Note:
        Either username/password or token/account_id must be provided.
        If token_file_path is provided, credentials will be saved/loaded
        automatically. When loading credentials, the current working directory
        and home directory are checked for the token file if no path is provided.
        When saving credentials, if no path is provided, it will save to the current
        working directory.
    """

    __slots__ = (
        '_account_id',
        '_country_code',
        '_current_region',
        '_password',
        '_token',
        '_token_file_path',
        '_username',
        'manager',
    )

    def __init__(
        self,
        manager: VeSync,
        username: str,
        password: str,
        country_code: str = DEFAULT_REGION,
    ) -> None:
        """Initialize VeSync Authentication Manager."""
        self.manager = manager
        self._username = username
        self._password = password
        self._token: str | None = None
        self._account_id: str | None = None
        self._country_code = country_code.upper()
        self._current_region = self._country_code_to_region()
        self._token_file_path: Path | None = None

    def _country_code_to_region(self) -> str:
        """Convert country code to region string for API use."""
        if self._country_code in NON_EU_COUNTRY_CODES:
            return 'US'
        return 'EU'

    @property
    def credentials_saved(self) -> bool:
        """Return whether credentials have been saved to file."""
        return self._token_file_path is not None and self._token_file_path.exists()

    @property
    def token(self) -> str:
        """Return VeSync API token."""
        if self._token is None:
            raise AttributeError(
                'Token not set, run VeSync.login or VeSync.set_token method'
            )
        return self._token

    @property
    def account_id(self) -> str:
        """Return VeSync account ID."""
        if self._account_id is None:
            raise AttributeError(
                'Account ID not set, run VeSync.login or VeSync.set_token method'
            )
        return self._account_id

    @property
    def country_code(self) -> str:
        """Return country code."""
        return self._country_code

    @country_code.setter
    def country_code(self, value: str) -> None:
        """Set country code."""
        self._country_code = value.upper()

    @property
    def current_region(self) -> str:
        """Return current region."""
        return self._current_region

    @property
    def is_authenticated(self) -> bool:
        """Check if user is authenticated."""
        return self._token is not None and self._account_id is not None

    def set_credentials(
        self,
        token: str,
        account_id: str,
        country_code: str,
        region: str,
    ) -> None:
        """Set authentication credentials.

        Args:
            token: Authentication token
            account_id: Account ID
            country_code: Country code in ISO 3166 Alpha-2 format
            region: Current region code
        """
        self._token = token
        self._account_id = account_id
        self._country_code = country_code.upper()
        self._current_region = region

    async def reauthenticate(self) -> bool:
        """Re-authenticate using stored username and password.

        Returns:
            True if re-authentication successful, False otherwise

        Raises:
            VeSyncLoginError: If login fails due to invalid credentials
            VeSyncAPIResponseError: If API response is invalid
            VeSyncServerError: If server returns an error
        """
        had_saved_credentials = self.credentials_saved
        self.clear_credentials()
        success = await self.login()
        logger.debug('Re-authentication successful for user: %s', self._username)
        if had_saved_credentials:
            await self.save_credentials_to_file(self._token_file_path)
        return success

    async def load_credentials_from_file(
        self, file_path: str | Path | None = None
    ) -> bool:
        """Load credentials from token file if path is set.

        If no path is provided, it will try to load from the current working directory and
        then the user's home directory.
        """
        locations = [Path.cwd() / '.vesync_auth', Path.home() / '.vesync_auth']
        file_path_object: Path | None = None
        if file_path is None:
            for location in locations:
                if location.exists():
                    file_path_object = location
                    break
        else:
            file_path_object = Path(file_path)
        if not file_path_object or not file_path_object.exists():
            logger.debug('Credentials file not found: %s', file_path_object)
            return False
        self._token_file_path = file_path_object
        try:
            data = await asyncio.to_thread(file_path_object.read_text, encoding='utf-8')
            data = orjson.loads(data)
            self._token = data['token']
            self._account_id = data['account_id']
            self._country_code = data['country_code'].upper()
            self._current_region = data['current_region'].upper()
            logger.debug('Credentials loaded from file: %s', file_path_object)
        except (orjson.JSONDecodeError, KeyError) as exc:
            logger.warning('Failed to load credentials from file: %s', exc)
            return False
        if self._token is None or self._account_id is None:
            logger.debug('Incomplete credentials in token file')
            self.manager.enabled = False
            return False
        self.manager.enabled = True
        return True

    def output_credentials_dict(self) -> dict[str, str] | None:
        """Output current credentials as a dictionary."""
        if self._token is None or self._account_id is None:
            logger.debug('No credentials to output, not authenticated')
            return None
        return {
            'token': self._token,
            'account_id': self._account_id,
            'country_code': self._country_code,
            'current_region': self._current_region,
        }

    def output_credentials_json(self) -> str | None:
        """Output current authentication credentials as a JSON string."""
        if not self.is_authenticated:
            logger.debug('No credentials to output, not authenticated')
            return None
        credentials = self.output_credentials_dict()
        if credentials is None:
            return None
        try:
            return orjson.dumps(credentials).decode('utf-8')
        except orjson.JSONEncodeError as exc:
            logger.warning('Failed to serialize credentials: %s', exc)
            return None

    async def save_credentials_to_file(self, file_path: str | Path | None = None) -> None:
        """Save authentication credentials to file."""
        if file_path is not None:
            file_path_object = Path(file_path)
        elif self._token_file_path is not None:
            file_path_object = self._token_file_path
        else:
            logger.debug('No token file path set, saving to default location')
            file_path_object = Path.cwd() / '.vesync_auth'
        if not self.is_authenticated:
            logger.debug('No credentials to save, not authenticated')
            return
        credentials = {
            'token': self._token,
            'account_id': self._account_id,
            'country_code': self._country_code,
            'current_region': self.current_region,
        }
        try:
            data = orjson.dumps(credentials).decode('utf-8')
            await asyncio.to_thread(file_path_object.write_text, data, encoding='utf-8')
            logger.debug('Credentials saved to file: %s', file_path_object)
        except (orjson.JSONEncodeError, OSError) as exc:
            logger.warning('Failed to save credentials to file: %s', exc)

    def clear_credentials(self) -> None:
        """Clear all stored credentials."""
        self._token = None
        self._account_id = None

        # Remove token file if it exists
        if self._token_file_path and self._token_file_path.exists():
            try:
                self._token_file_path.unlink()
                logger.debug('Token file deleted: %s', self._token_file_path)
            except OSError as exc:
                logger.warning('Failed to delete token file: %s', exc)

    async def login(self) -> bool:
        """Log into VeSync server using username/password or existing token.

        Returns:
            True if login successful, False otherwise

        Raises:
            VeSyncLoginError: If login fails due to invalid credentials
            VeSyncAPIResponseError: If API response is invalid
            VeSyncServerError: If server returns an error
        """
        # Attempt username/password login
        if self._username and self._password:
            return await self._login_with_credentials()

        raise VeSyncLoginError(
            'No valid authentication method available. '
            'Provide either username/password or valid token/account_id.'
        )

    async def _login_with_credentials(self) -> bool:
        """Login using username and password.

        Returns:
            True if login successful

        Raises:
            VeSyncLoginError: If login fails
            VeSyncAPIResponseError: If API response is invalid
        """
        if not isinstance(self._username, str) or len(self._username) == 0:
            raise VeSyncLoginError('Username must be a non-empty string')

        if not isinstance(self._password, str) or len(self._password) == 0:
            raise VeSyncLoginError('Password must be a non-empty string')

        try:
            # Step 1: Get authorization code
            auth_code = await self._get_authorization_code()

            # Step 2: Exchange authorization code for token
            await self._exchange_authorization_code(auth_code)
            logger.debug('Login successful for user: %s', self._username)
        except (VeSyncLoginError, VeSyncAPIResponseError, VeSyncServerError):
            raise
        except Exception as exc:
            msg = f'Unexpected error during login: {exc}'
            raise VeSyncLoginError(msg) from exc
        return True

    async def _get_authorization_code(self) -> str:
        """Get authorization code using username and password.

        Returns:
            Authorization code

        Raises:
            VeSyncAPIResponseError: If API response is invalid
            VeSyncLoginError: If authentication fails
        """
        request_auth = RequestGetTokenModel(
            email=self._username,  # type: ignore[arg-type]
            method='authByPWDOrOTM',
            password=self._password,  # type: ignore[arg-type]
        )

        resp_dict, _ = await self.manager.async_call_api(
            '/globalPlatform/api/accountAuth/v1/authByPWDOrOTM',
            'post',
            json_object=request_auth,
        )

        if resp_dict is None:
            raise VeSyncAPIResponseError('Error receiving response to auth request')

        if resp_dict.get('code') != 0:
            error_info = ErrorCodes.get_error_info(resp_dict.get('code'))
            resp_message = resp_dict.get('msg')

            if resp_message is not None:
                error_info.message = f'{error_info.message} ({resp_message})'

            msg = f'Authentication failed - {error_info.message}'
            raise VeSyncLoginError(msg)

        try:
            response_model = ResponseLoginModel.from_dict(resp_dict)
        except UnserializableDataError as exc:
            logger.debug('Error parsing auth response: %s', exc)
            raise VeSyncAPIResponseError('Error parsing authentication response') from exc
        except MissingField as exc:
            logger.debug('Error parsing auth response: %s', exc)
            raise VeSyncAPIResponseError('Error parsing authentication response') from exc

        result = response_model.result
        if not isinstance(result, RespGetTokenResultModel):
            raise VeSyncAPIResponseError('Invalid authentication response format')
        self._account_id = result.accountID
        return result.authorizeCode

    async def _exchange_authorization_code(
        self,
        auth_code: str,
        region_change_token: str | None = None,
    ) -> None:
        """Exchange authorization code for access token.

        Args:
            auth_code: Authorization code from first auth step
            region_change_token: Token for region change (retry scenario)

        Raises:
            VeSyncLoginError: If login fails
            VeSyncAPIResponseError: If API response is invalid
        """
        request_login = RequestLoginTokenModel(
            method='loginByAuthorizeCode4Vesync',
            authorizeCode=auth_code,
            bizToken=region_change_token,
            userCountryCode=self._country_code,
            regionChange='lastRegion' if region_change_token else None,
        )

        resp_dict, _ = await self.manager.async_call_api(
            '/user/api/accountManage/v1/loginByAuthorizeCode4Vesync',
            'post',
            json_object=request_login,
        )

        if resp_dict is None:
            raise VeSyncAPIResponseError('Error receiving response to login request')

        try:
            response_model = ResponseLoginModel.from_dict(resp_dict)

            if not isinstance(response_model.result, RespLoginTokenResultModel):
                raise VeSyncAPIResponseError('Invalid login response format')

            if response_model.code != 0:
                error_info = ErrorCodes.get_error_info(resp_dict.get('code'))

                # Handle cross region error by retrying with new region
                if error_info.error_type == ErrorTypes.CROSS_REGION:
                    result = response_model.result
                    self._country_code = result.countryCode
                    self._current_region = result.currentRegion
                    logger.debug(
                        'Cross-region error, retrying with country: %s',
                        self._country_code,
                    )
                    return await self._exchange_authorization_code(
                        auth_code, region_change_token=result.bizToken
                    )

                resp_message = resp_dict.get('msg')
                if resp_message is not None:
                    error_info.message = f'{error_info.message} ({resp_message})'

                msg = f'Login failed - {error_info.message}'

                # Raise appropriate exception based on error type
                if error_info.error_type == ErrorTypes.SERVER_ERROR:
                    raise VeSyncServerError(msg)
                raise VeSyncLoginError(msg)

            result = response_model.result
            self._token = result.token
            self._account_id = result.accountID
            self._country_code = result.countryCode

        except (MissingField, UnserializableDataError) as exc:
            logger.debug('Error parsing login response: %s', exc)
            raise VeSyncAPIResponseError('Error parsing login response') from exc

    def __repr__(self) -> str:
        """Return string representation of VeSyncAuth."""
        return (
            f'VeSyncAuth(username={self._username!r}, '
            f'country_code={self._country_code!r}, '
            f'authenticated={self.is_authenticated})'
        )



"""Data models for VeSync Fans.

These models inherit from `ResponseBaseModel` and `RequestBaseModel` from the
`base_models` module.
"""

from __future__ import annotations

from dataclasses import dataclass

from mashumaro.mixins.orjson import DataClassORJSONMixin

from pyvesync.models.bypass_models import (
    BypassV2InnerResult,
)


@dataclass
class TowerFanResult(BypassV2InnerResult):
    """Tower Fan Result Model."""

    powerSwitch: int
    workMode: str
    manualSpeedLevel: int
    fanSpeedLevel: int
    screenState: int
    screenSwitch: int
    oscillationSwitch: int
    oscillationState: int
    muteSwitch: int
    muteState: int
    timerRemain: int
    temperature: int
    errorCode: int
    scheduleCount: int
    displayingType: int | None = None
    sleepPreference: FanTowerSleepPreferences | None = None


@dataclass
class FanSleepPreferences(DataClassORJSONMixin):
    """Fan Sleep Preferences."""


@dataclass
class FanTowerSleepPreferences(FanSleepPreferences):
    """Fan Sleep Preferences."""

    sleepPreferenceType: str
    oscillationSwitch: int
    fallAsleepRemain: int
    autoChangeFanLevelSwitch: int


@dataclass
class FanPedestalSleepPreferences(FanSleepPreferences):
    """Fan Sleep Preferences."""

    sleepPreferenceType: str
    oscillationState: int
    fallAsleepRemain: int
    initFanSpeedLevel: int


@dataclass
class PedestalFanResult(BypassV2InnerResult):
    """Pedestal Fan Result Model."""

    powerSwitch: int
    workMode: str
    fanSpeedLevel: int
    temperature: int
    muteSwitch: int
    muteState: int
    screenState: int
    screenSwitch: int
    horizontalOscillationState: int
    verticalOscillationState: int
    childLock: int
    errorCode: int
    oscillationCoordinate: OscillationCoordinatesModel | None = None
    oscillationRange: OscillationRangeModel | None = None
    sleepPreference: FanPedestalSleepPreferences | None = None


@dataclass
class OscillationCoordinatesModel(DataClassORJSONMixin):
    """Oscillation Coordinates Model."""

    yaw: int
    pitch: int


@dataclass
class OscillationRangeModel(DataClassORJSONMixin):
    """Oscillation Range Model."""

    left: int
    right: int
    top: int
    bottom: int



"""Mixins for Devices that have similar API requests.

The `BypassV2Mixin` and `BypassV1Mixin` classes are used to send API requests to the
Bypass V2 and Bypass V1 endpoints, respectively. The `BypassV2Mixin` class is used for
devices that use the `/cloud/v2/deviceManaged/bypassV2` endpoint, while the
`BypassV1Mixin` class is used for devices that use the
`/cloud/v1/deviceManaged/{endpoint}` path.
"""

from __future__ import annotations

from logging import Logger
from typing import TYPE_CHECKING, TypeVar

from mashumaro.mixins.orjson import DataClassORJSONMixin

from pyvesync.models.bypass_models import (
    RequestBypassV1,
    RequestBypassV2,
)
from pyvesync.utils.helpers import Helpers
from pyvesync.utils.logs import LibraryLogger

if TYPE_CHECKING:
    from pyvesync import VeSync
    from pyvesync.base_devices import VeSyncBaseDevice

T_MODEL = TypeVar('T_MODEL', bound=DataClassORJSONMixin)

BYPASS_V1_PATH = '/cloud/v1/deviceManaged/'
BYPASS_V2_BASE = '/cloud/v2/deviceManaged/'


def process_bypassv1_result(
    device: VeSyncBaseDevice,
    logger: Logger,
    method: str,
    resp_dict: dict | None,
    model: type[T_MODEL],
) -> T_MODEL | None:
    """Process the Bypass V1 API response.

    This will gracefully handle errors in the response and error codes,
    logging them as needed. The return value is the built model.

    Args:
        device (VeSyncBaseDevice): The device object.
        logger (Logger): The logger to use for logging.
        method (str): The method used in the payload.
        resp_dict (dict | str): The api response.
        model (type[T_MODEL]): The model to use for the response.

    Returns:
        dict: The response data
    """
    r_dict = Helpers.process_dev_response(logger, method, device, resp_dict)
    if r_dict is None or 'result' not in r_dict:
        return None
    result = r_dict['result']
    return Helpers.model_maker(logger, model, method, result, device)


def _get_inner_result(
    device: VeSyncBaseDevice,
    logger: Logger,
    method: str,
    resp_dict: dict,
) -> dict | None:
    """Process the code in the result field of Bypass V2."""
    try:
        outer_result = resp_dict['result']
        inner_result = outer_result['result']
        code = int(outer_result['code'])
    except (ValueError, TypeError, KeyError):
        LibraryLogger.error_device_response_content(
            logger,
            device,
            method,
            'Error processing bypass V2 API response result.',
        )
        return None

    if code != 0:
        return None
    return inner_result


def process_bypassv2_result(
    device: VeSyncBaseDevice,
    logger: Logger,
    method: str,
    resp_dict: dict | None,
    model: type[T_MODEL],
) -> T_MODEL | None:
    """Process the Bypass V1 API response.

    This will gracefully handle errors in the response and error codes,
    logging them as needed. The return dictionary is the **inner** result value of
    the API response.

    Args:
        device (VeSyncBaseDevice): The device object.
        logger (Logger): The logger to use for logging.
        method (str): The method used in the payload.
        resp_dict (dict | str): The api response.
        model (type[T_MODEL]): The model to use for the response.

    Returns:
        T_MODEL: An instance of the inner result model.
    """
    r_dict = Helpers.process_dev_response(logger, method, device, resp_dict)
    if r_dict is None:
        return None
    result = _get_inner_result(device, logger, method, r_dict)
    if not isinstance(result, dict):
        return None
    return Helpers.model_maker(logger, model, method, result, device)


class BypassV2Mixin:
    """Mixin for bypass V2 API.

    Overrides the `_build_request` method and `request_keys` attribute for devices
    that use the Bypass V2 API- /cloud/v2/deviceManaged/bypassV2.
    """

    if TYPE_CHECKING:
        manager: VeSync

    __slots__ = ()
    request_keys: tuple[str, ...] = (
        'acceptLanguage',
        'appVersion',
        'phoneBrand',
        'phoneOS',
        'accountID',
        'cid',
        'configModule',
        'debugMode',
        'traceId',
        'timeZone',
        'token',
        'userCountryCode',
        'configModel',
        'deviceId',
    )

    def _build_request(
        self,
        payload_method: str,
        data: dict | None = None,
        method: str = 'bypassV2',
        payload_update: dict | None = None,
    ) -> RequestBypassV2:
        """Build API request body Bypass V2 endpoint.

        Args:
            payload_method (str): The method to use in the payload dict.
            data (dict | None): The data dict inside the payload value.
            method (str): The method to use in the outer body, defaults to bypassV2.
            payload_update (dict | None): Additional keys to add to the payload.
        """
        body = {}
        body = Helpers.get_defaultvalues_attributes(self.request_keys).copy()
        body.update(Helpers.get_manager_attributes(self.manager, self.request_keys))
        body.update(Helpers.get_device_attributes(self, self.request_keys))
        body['method'] = method
        body['payload'] = {'method': payload_method, 'source': 'APP', 'data': data or {}}
        body['payload'] |= payload_update or {}
        return RequestBypassV2.from_dict(body)

    async def call_bypassv2_api(
        self,
        payload_method: str,
        data: dict | None = None,
        method: str = 'bypassV2',
        endpoint: str = 'bypassV2',
        payload_update: dict | None = None,
    ) -> dict | None:
        """Send Bypass V2 API request.

        This uses the `_build_request` method to send API requests to the Bypass V2 API.

        Args:
            payload_method (str): The method to use in the payload dict.
            data (dict | None): The data to send in the request.
            method (str): The method to use in the outer body.
            endpoint (str | None): The last part of the API url, defaults to
                `bypassV2`, e.g. `/cloud/v2/deviceManaged/bypassV2`.
            payload_update (dict | None): Additional keys to add to the payload.

        Returns:
            bytes: The response from the API request.
        """
        request = self._build_request(payload_method, data, method, payload_update)
        endpoint = BYPASS_V2_BASE + endpoint
        resp_dict, _ = await self.manager.async_call_api(
            endpoint, 'post', request, Helpers.req_header_bypass()
        )
        return resp_dict


class BypassV1Mixin:
    """Mixin for bypass V1 API.

    Overrides the `_build_request` method and `request_keys` attribute for devices
    that use the Bypass V1 API- /cloud/v1/deviceManaged/bypass. The primary method to
    call is `call_bypassv1_api`, which is a wrapper for setting up the request body and
    calling the API. The `bypass` endpoint can also be overridden for specific API calls.
    """

    if TYPE_CHECKING:
        manager: VeSync

    __slots__ = ()
    request_keys: tuple[str, ...] = (
        'acceptLanguage',
        'appVersion',
        'phoneBrand',
        'phoneOS',
        'accountID',
        'cid',
        'configModule',
        'debugMode',
        'traceId',
        'timeZone',
        'token',
        'userCountryCode',
        'uuid',
        'configModel',
        'deviceId',
    )

    def _build_request(
        self,
        request_model: type[RequestBypassV1],
        update_dict: dict | None = None,
        method: str = 'bypass',
    ) -> RequestBypassV1:
        """Build API request body for the Bypass V1 endpoint.

        Args:
            request_model (type[RequestBypassV1]): The request model to use.
            update_dict (dict | None): Additional keys to add on.
            method (str): The method to use in the outer body, defaults to bypass.

        Returns:
            RequestBypassV1: The request body for the Bypass V1 endpoint, the correct
            model is determined from the RequestBypassV1 discriminator.
        """
        body = Helpers.get_defaultvalues_attributes(self.request_keys).copy()
        body.update(Helpers.get_manager_attributes(self.manager, self.request_keys))
        body.update(Helpers.get_device_attributes(self, self.request_keys))
        body['method'] = method
        body.update(update_dict or {})
        return request_model.from_dict(body)

    async def call_bypassv1_api(
        self,
        request_model: type[RequestBypassV1],
        update_dict: dict | None = None,
        method: str = 'bypass',
        endpoint: str = 'bypass',
    ) -> dict | None:
        """Send ByPass V1 API request.

        This uses the `_build_request` method to send API requests to the Bypass V1 API.
        The endpoint can be overridden with the `endpoint` argument.

        Args:
            request_model (type[RequestBypassV1]): The request model to use.
            update_dict (dict): Additional keys to add on.
            method (str): The method to use in the outer body.
            endpoint (str | None): The last part of the url path, defaults to
                `bypass`, e.g. `/cloud/v1/deviceManaged/bypass`.

        Returns:
            bytes: The response from the API request.
        """
        request = self._build_request(request_model, update_dict, method)
        url_path = BYPASS_V1_PATH + endpoint
        resp_dict, _ = await self.manager.async_call_api(
            url_path, 'post', request, Helpers.req_header_bypass()
        )

        return resp_dict



"""Data models for VeSync Humidifier devices.

These models inherit from `ResponseBaseModel` and `RequestBaseModel` from the
`base_models` module.

The `InnerHumidifierBaseResult` class is used as a base class for the inner humidifier
result models. The correct subclass is determined by the mashumaro discriminator.
"""

from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated

from mashumaro.config import BaseConfig
from mashumaro.types import Alias

from pyvesync.models.base_models import (
    ResponseBaseModel,
    ResponseCodeModel,
)


@dataclass
class ResponseHumidifierBase(ResponseCodeModel):
    """Humidifier Base Response Dict."""

    result: OuterHumidifierResult


@dataclass
class OuterHumidifierResult(ResponseBaseModel):
    """Humidifier Result Dict."""

    code: int
    result: InnerHumidifierBaseResult


@dataclass
class InnerHumidifierBaseResult(ResponseBaseModel):
    """Base class for inner humidifier results model.

    All inner results models inherit from this class and are
    correctly subclassed by the mashumaro discriminator.
    """

    class Config(BaseConfig):  # type: ignore[override]
        """Configure the results model to use subclass discriminator."""

        allow_deserialization_not_by_alias = True


class BypassV2InnerErrorResult(InnerHumidifierBaseResult):
    """Inner Error Result Model."""

    msg: str


# Inner Result models for individual devices inherit from InnerHumidifierBaseResult
# and are used to parse the response from the API.
# The correct subclass is determined by the mashumaro discriminator


@dataclass
class ClassicLVHumidResult(InnerHumidifierBaseResult):
    """Classic 200S Humidifier Result Model.

    Inherits from InnerHumidifierBaseResult.
    """

    enabled: bool
    mist_virtual_level: int
    mist_level: int
    mode: str
    display: Annotated[bool, Alias('indicator_light_switch')]
    water_lacks: bool
    humidity: int | None = None
    humidity_high: bool = False
    automatic_stop_reach_target: bool = False
    water_tank_lifted: bool = False
    warm_enabled: bool = False
    warm_level: int | None = None
    night_light_brightness: int | None = None
    configuration: ClassicConfig | None = None


@dataclass
class ClassicConfig(ResponseBaseModel):
    """Classic 200S Humidifier Configuration Model."""

    auto_target_humidity: int = 0
    display: Annotated[bool, Alias('indicator_light_switch')] = False
    automatic_stop: bool = False

    class Config(BaseConfig):  # type: ignore[override]
        """Configure the results model to use subclass discriminator."""

        allow_deserialization_not_by_alias = True
        forbid_extra_keys = False


@dataclass
class LV600SConfig(ResponseBaseModel):
    """LV 600S Humidifier Configuration Model."""

    auto_target_humidity: int = 0
    display: bool = False


@dataclass
class LV600SExtension(ResponseBaseModel):
    """LV 600S Humidifier Configuration Model."""

    timer_remain: int = 0
    schedule_count: int = 0


@dataclass
class LV600SHumidResult(InnerHumidifierBaseResult):
    """LV600S Humidifier Result Model.

    Inherits from InnerHumidifierBaseResult.
    """

    automatic_stop_reach_target: bool
    display: bool
    enabled: bool
    humidity: int
    humidity_high: bool
    mist_level: int
    mist_virtual_level: int
    mode: str
    water_lacks: bool
    water_tank_lifted: bool
    extension: LV600SExtension | None = None
    configuration: LV600SConfig | None = None


# Models for the VeSync Superior 6000S Humidifier


@dataclass
class Superior6000SResult(InnerHumidifierBaseResult):
    """Superior 6000S Humidifier Result Model.

    Inherits from InnerHumidifierBaseResult.
    """

    powerSwitch: int
    humidity: int
    targetHumidity: int
    virtualLevel: int
    mistLevel: int
    workMode: str
    waterLacksState: int
    waterTankLifted: int
    autoStopSwitch: int
    autoStopState: int
    screenSwitch: int
    screenState: int
    scheduleCount: int
    timerRemain: int
    errorCode: int
    autoPreference: int
    childLockSwitch: int
    filterLifePercent: int
    temperature: int
    dryingMode: DryingModeModel | None = None


@dataclass
class DryingModeModel(ResponseBaseModel):
    """Drying Mode Model for Superior 6000S Humidifier."""

    dryingLevel: int
    autoDryingSwitch: int
    dryingState: int
    dryingRemain: int


# Models for the Levoit 1000S Humidifier


@dataclass
class Levoit1000SResult(InnerHumidifierBaseResult):
    """Levoit 1000S Humidifier Result Model."""

    powerSwitch: int
    humidity: int
    targetHumidity: int
    virtualLevel: int
    mistLevel: int
    workMode: str
    waterLacksState: int
    waterTankLifted: int
    autoStopSwitch: int
    autoStopState: int
    screenSwitch: int
    screenState: int
    scheduleCount: int
    timerRemain: int
    errorCode: int
    nightLight: Levoit1000SNightLight | None = None


@dataclass
class Levoit1000SNightLight(ResponseBaseModel):
    """Night Light Model for Levoit 1000S Humidifier."""

    nightLightSwitch: int
    brightness: int


@dataclass
class SproutHumidifierResult(InnerHumidifierBaseResult):
    """Sprout Humidifier Result Model."""

    powerSwitch: int
    humidity: int
    targetHumidity: int
    virtualLevel: int
    mistLevel: int
    workMode: str
    waterLacksState: int
    waterTankLifted: int
    autoStopSwitch: int
    autoStopState: int
    screenSwitch: int
    screenState: int
    autoModePreference: int
    autoPreference: int
    waterShortageDryingSwitch: int
    childLockSwitch: int
    filterLifePercent: int
    hepaFilterLifePercent: int
    temperature: int
    lampSwitch: int
    lampType: int
    dumpedState: int
    roomSize: int
    supportLampAct: int
    lastDryingCompletedTime: int
    afterDryLastHumidityTime: int
    sensorContent: SproutSensorcontent | None = None
    breathingLamp: SproutBreathinglamp | None = None
    guardingInfo: SproutGuardinginfo | None = None
    dryingMode: DryingModeModel | None = None
    nightLight: SproutNightlight | None = None


@dataclass
class SproutSensorcontent(ResponseBaseModel):
    """Sprout Humidifier Sensor Content Model."""

    selfSensorMac: str
    selfSenorBattery: int
    selfSensorStatus: str


@dataclass
class SproutNightlight(ResponseBaseModel):
    """Sprout Humidifier Night Light Model."""

    nightLightSwitch: int
    brightness: int
    colorTemperature: int


@dataclass
class SproutBreathinglamp(ResponseBaseModel):
    """Sprout Humidifier Breathing Lamp Model."""

    breathingLampSwitch: int
    colorTemperature: int
    timeInterval: int
    brightnessStart: int
    brightnessEnd: int


@dataclass
class SproutGuardinginfo(ResponseBaseModel):
    """Sprout Humidifier Guarding Info Model."""

    guarding: int
    remainTS: int


@dataclass
class LV600SResult(InnerHumidifierBaseResult):
    """LV600S (LUH-A603S) Humidifier Result Model.

    Uses the newer API format with powerSwitch, workMode, etc.
    Includes warm mist support via warmPower and warmLevel.
    """

    powerSwitch: int
    humidity: int
    targetHumidity: int
    virtualLevel: int
    mistLevel: int
    workMode: str
    waterLacksState: int
    waterTankLifted: int
    autoStopSwitch: int
    autoStopState: int
    screenSwitch: int
    screenState: int
    scheduleCount: int
    timerRemain: int
    errorCode: int
    totalWorkTime: int = 0
    warmPower: bool = False
    warmLevel: int = 0



"""Get configuration details from the API."""

from __future__ import annotations

import logging
from pathlib import Path
from typing import Any

import aiofiles
import orjson

from pyvesync import VeSync
from pyvesync.models.vesync_models import RequestDeviceConfiguration

USERNAME = ''
PASSWORD = ''

vs_logger = logging.getLogger('pyvesync')
vs_logger.setLevel(logging.DEBUG)

logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)


def parse_config(data: dict) -> dict[str, list[dict]]:
    """Parse the configuration data from the API into a nested dictionary."""
    result = {}
    # Navigate to productLineList
    data = orjson.loads(orjson.dumps(data))  # Ensure data is a dict
    config_list = data.get('result', {}).get('configList', [])
    for config in config_list:
        for item in config.get('items', []):
            item_value = item.get('itemValue')
            if isinstance(item_value, str):
                item_value = orjson.loads(item_value)

            product_lines = item_value.get('productLineList', [])
            for product_line in product_lines:
                for type_info in product_line.get('typeInfoList', []):
                    type_name = type_info.get('typeName')
                    if not type_name:
                        continue
                    models = []
                    for model_info in type_info.get('modelInfoList', []):
                        model = model_info.get('model')
                        if not model:
                            continue
                        # Add the full model_info dict under the model key
                        models.append({model: [model_info]})
                    if models:
                        result[type_name] = models
    return result


async def fetch_config(manager: VeSync) -> dict[Any, Any] | None:
    """Get full device configuration from the API."""
    endpoint = '/cloud/v1/app/getAppConfigurationV2'
    body = RequestDeviceConfiguration(
        accountID=manager.account_id,
        token=manager.token,
    )
    response, _ = await manager.async_call_api(
        api=endpoint,
        method='post',
        json_object=body,
    )
    return response


async def main() -> None:
    """Main function to fetch and display device configurations."""
    async with VeSync(USERNAME, PASSWORD, 'US') as manager:
        await manager.login()
        config = await fetch_config(manager)
        if not config:
            print('Failed to fetch configuration.')
            return
        parsed_config = parse_config(config)
        print(orjson.dumps(parsed_config, option=orjson.OPT_INDENT_2).decode('utf-8'))
        output_path = Path('models.json')
        async with aiofiles.open(output_path, 'w', encoding='utf-8') as file:
            await file.write(
                orjson.dumps(parsed_config, option=orjson.OPT_INDENT_2).decode('utf-8')
            )


if __name__ == '__main__':
    import asyncio

    asyncio.run(main())



"""Base devices and state classes for VeSync devices.

This module contains the base classes for VeSync devices, as well as the state classes
for each device type. The base classes are used to create the device objects, while the
state classes are used to store the current state of the device.
"""

from .bulb_base import BulbState, VeSyncBulb
from .fan_base import FanState, VeSyncFanBase
from .fryer_base import FryerState, VeSyncFryer
from .humidifier_base import HumidifierState, VeSyncHumidifier
from .outlet_base import OutletState, VeSyncOutlet
from .purifier_base import PurifierState, VeSyncPurifier
from .switch_base import SwitchState, VeSyncSwitch
from .vesyncbasedevice import DeviceState, VeSyncBaseDevice, VeSyncBaseToggleDevice

__all__ = [
    'BulbState',
    'DeviceState',
    'FanState',
    'FryerState',
    'HumidifierState',
    'OutletState',
    'PurifierState',
    'SwitchState',
    'VeSyncBaseDevice',
    'VeSyncBaseToggleDevice',
    'VeSyncBulb',
    'VeSyncFanBase',
    'VeSyncFryer',
    'VeSyncHumidifier',
    'VeSyncOutlet',
    'VeSyncPurifier',
    'VeSyncSwitch',
]



"""Data models for VeSync API requests and responses.

Models should follow the naming convention of Request/Response + API Name.
Internal models can have any descriptive name.

The `base_models` module contains the base classes for the VeSync API models, while the
models for each device type and other API calls are stored in their respective modules.

The [bypassv2_models][pyvesync.models.bypass_models] module contains the models for
the common bypassV2 API calls.
"""



"""Module holding VeSync API Error Codes and Response statuses.

Error codes are pulled from the VeSync app source code. If there are unknown errors
that are not found, please open an issue on GitHub.

Errors are integers divided by 1000 and then multiplied by 1000 to
get the base error code. It also tries to check if the absolute value matches
as well.

This is used by the `pyvesync.utils.helpers.Helpers.process_dev_response` method to
retrieve response code information and store in the `last_response` device instance.


The "check_device" key of the error dictionary is used to determine if the logger
should emit a warning to the user for critical device errors, such as a short
or voltage error.

The ResponseInfo class is used to define information regarding the error code or
indicate the request was successful. The ErrorTypes class is used to categorize the
error types. This is a WIP and subject to change. Should not be used for anything
other than the ErrorCodes error configuration.

Example:
    Example of the error dictionary structure:
    ```python
    pyvesync.errors.ErrorCodes.get_error_info("-11201022")
    ResponseInfo(
        name="PASSWORD_ERROR",
        error_type=ErrorTypes.AUTHENTICATION,
        msg="Invalid password"
        critical_error=False,
        operational_error=False,
        device_online=None
    )

    device.last_response
    # Returns the ResponseInfo object.
    ```
"""

from __future__ import annotations

from dataclasses import dataclass, replace
from enum import StrEnum
from types import MappingProxyType

from mashumaro.mixins.orjson import DataClassORJSONMixin


@dataclass
class ResponseInfo(DataClassORJSONMixin):
    """Class holding response information and error code definitions and lookup methods.

    Attributes:
        name (str): Name of the error
        error_type (ErrorTypes): Type of the error see `ErrorTypes`
        message (str): Message for the error
        critical_error (bool): A major error, such as a short or voltage error
        operational_error (bool): Device connected but API error
        device_online (bool | None): Device online status
        code (str | int | None): Error code from API
        response_data (dict | None): Response data from API - populated by the
            process_dev_response method in the Helpers class and bypass mixin
            methods.
    """

    name: str
    error_type: str
    message: str
    critical_error: bool = False
    operational_error: bool = False  # Device connected but API error
    device_online: bool = True  # Defaults to connected
    code: int | None = None
    response_data: dict | None = None  # Response data from API


class ErrorTypes(StrEnum):
    """Error types for API return codes.

    Attributes:
        SUCCESS: Successful request
        AUTHENTICATION: Authentication error
        RATE_LIMIT: Rate limiting error
        SERVER_ERROR: Server and connection error
        REQUEST_ERROR: Error in Request parameters or method
        DEVICE_ERROR: Device operational error, device
            connected but cannot perform requested action
        CONFIG_ERROR: Configuration error in user profile
        DEVICE_OFFLINE: Device is offline, not connected
        UNKNOWN_ERROR: Unknown error
        BAD_RESPONSE: Bad response from API
    """

    SUCCESS = 'success'
    AUTHENTICATION = 'auth_error'
    RATE_LIMIT = 'rate_limit_error'
    SERVER_ERROR = 'server_error'
    REQUEST_ERROR = 'request_error'
    DEVICE_ERROR = 'device_error'
    CONFIG_ERROR = 'config_error'
    DEVICE_OFFLINE = 'device_offline'
    UNKNOWN_ERROR = 'unknown_error'
    TOKEN_ERROR = 'token_error'
    BAD_RESPONSE = 'bad_response'
    CROSS_REGION = 'cross_region'


class ErrorCodes:
    """Class holding error code definitions and lookup methods.

    Taken from VeSync app source code. Values are `ErrorInfo` dataclasses.

    Attributes:
        errors (MappingProxyType[str, ErrorInfo]): Dictionary of error codes and
            their meanings.

    Example:
        Error codes are taken from VeSync app source code, if there are unknown errors
        that are not found, please open an issue on GitHub. The "critical_error" key of
        the error dictionary is used to determine if the logger should emit a warning.
        Used for device issues that are critical, such as a short or voltage error.

        Each error dictionary is structured as follows:
        ```
        ErrorInfo(
            name: str,
            error_type: ErrorType,
            message: str,
            critical_error: bool,
            operational_error: bool,
            device_online: bool
            )
        ```

        The `cls.critical_error(error_code)` method is used to determine if the error
        is a critical device error that should emit a warning.

        `cls.get_error_info(error_code)` method is used to return the error dictionary
        for a given error code.
    """

    errors: MappingProxyType[str, ResponseInfo] = MappingProxyType(
        {
            '-11260022': ResponseInfo(
                'CROSS_REGION_ERROR',
                ErrorTypes.CROSS_REGION,
                'Cross region error',
            ),
            '11': ResponseInfo(
                'DEVICE_OFFLINE',
                ErrorTypes.DEVICE_OFFLINE,
                'Device offline',
                device_online=False,
            ),
            '4041004': ResponseInfo(
                'DEVICE_OFFLINE',
                ErrorTypes.DEVICE_OFFLINE,
                'Device offline',
                device_online=False,
            ),
            '-11203000': ResponseInfo(
                'ACCOUNT_EXIST', ErrorTypes.AUTHENTICATION, 'Account already exists'
            ),
            '-11200000': ResponseInfo(
                'ACCOUNT_FORMAT_ERROR', ErrorTypes.CONFIG_ERROR, 'Account format error'
            ),
            '-11202000': ResponseInfo(
                'ACCOUNT_NOT_EXIST', ErrorTypes.AUTHENTICATION, 'Account does not exist'
            ),
            '-11300027': ResponseInfo(
                'AIRPURGE_OFFLINE',
                ErrorTypes.DEVICE_OFFLINE,
                'Device offline',
                device_online=False,
            ),
            '-16906000': ResponseInfo(
                'REQUEST_TOO_FREQUENT',
                ErrorTypes.RATE_LIMIT,
                'Request too frequent',
                operational_error=True,
            ),
            '-11902000': ResponseInfo(
                'AUTHKEY_NOT_EXIST', ErrorTypes.CONFIG_ERROR, 'Authkey does not exist'
            ),
            '-11900000': ResponseInfo(
                'AUTHKEY_PID_NOT_MATCH', ErrorTypes.DEVICE_ERROR, 'Authkey PID mismatch'
            ),
            '-11504000': ResponseInfo(
                'AWAY_MAX', ErrorTypes.CONFIG_ERROR, 'Away maximum reached'
            ),
            '11014000': ResponseInfo(
                'BYPASS_AIRPURIFIER_E2',
                ErrorTypes.DEVICE_ERROR,
                'Air Purifier E2 error',
                critical_error=True,
                device_online=True,
            ),
            '11802000': ResponseInfo(
                'BYPASS_AIRPURIFIER_MOTOR_ABNORMAL',
                ErrorTypes.DEVICE_ERROR,
                'Air Purifier motor error',
                critical_error=True,
                device_online=True,
            ),
            '11504000': ResponseInfo(
                'BYPASS_AWAY_MAX', ErrorTypes.CONFIG_ERROR, 'Away maximum reached'
            ),
            '11509000': ResponseInfo(
                'BYPASS_AWAY_NOT_EXIST',
                ErrorTypes.CONFIG_ERROR,
                'Away does not exist',
            ),
            '11908000': ResponseInfo(
                'BYPASS_COOK_TIMEOUT', ErrorTypes.DEVICE_ERROR, 'Cook timeout error'
            ),
            '11909000': ResponseInfo(
                'BYPASS_SMART_STOP',
                ErrorTypes.DEVICE_ERROR,
                'Smart stop error',
                device_online=True,
            ),
            '11910000': ResponseInfo(
                'BYPASS_LEFT_ZONE_COOKING',
                ErrorTypes.DEVICE_ERROR,
                'Left zone cooking error',
                device_online=True,
            ),
            '11911000': ResponseInfo(
                'BYPASS_RIGHT_ZONE_COOKING',
                ErrorTypes.DEVICE_ERROR,
                'Right zone cooking error',
                device_online=True,
            ),
            '11912000': ResponseInfo(
                'BYPASS_ALL_ZONE_COOKING',
                ErrorTypes.DEVICE_ERROR,
                'All zone cooking error',
                device_online=True,
            ),
            '11916000': ResponseInfo(
                'BYPASS_NTC_RIGHT_TOP_SHORT',
                ErrorTypes.DEVICE_ERROR,
                'Right top short error',
                critical_error=True,
                device_online=True,
            ),
            '11917000': ResponseInfo(
                'BYPASS_NTC_RIGHT_TOP_OPEN',
                ErrorTypes.DEVICE_ERROR,
                'Right top open error',
                critical_error=True,
                device_online=True,
            ),
            '11918000': ResponseInfo(
                'BYPASS_NTC_BOTTOM_SHORT',
                ErrorTypes.DEVICE_ERROR,
                'Bottom short error',
                critical_error=True,
                device_online=True,
            ),
            '11919000': ResponseInfo(
                'BYPASS_NTC_BOTTOM_OPEN',
                ErrorTypes.DEVICE_ERROR,
                'Bottom open error',
                critical_error=True,
                device_online=True,
            ),
            '11924000': ResponseInfo(
                'BYPASS_RIGHT_TEMP_FAULT',
                ErrorTypes.DEVICE_ERROR,
                'Right temperature fault',
                critical_error=True,
                device_online=True,
            ),
            '11925000': ResponseInfo(
                'BYPASS_ZONE_2_MOTOR_ABNORMAL',
                ErrorTypes.DEVICE_ERROR,
                'Zone 2 motor error',
                critical_error=True,
                device_online=True,
            ),
            '11021000': ResponseInfo(
                'BYPASS_DEVICE_END',
                ErrorTypes.DEVICE_ERROR,
                'Device end error',
                critical_error=True,
                device_online=True,
            ),
            '11012000': ResponseInfo(
                'BYPASS_DEVICE_RUNNING',
                ErrorTypes.DEVICE_ERROR,
                'Device running error',
                critical_error=True,
                device_online=True,
            ),
            '11020000': ResponseInfo(
                'BYPASS_DEVICE_STOP',
                ErrorTypes.DEVICE_ERROR,
                'Device stop error',
                device_online=True,
                critical_error=True,
            ),
            '11901000': ResponseInfo(
                'BYPASS_DOOR_OPEN',
                ErrorTypes.DEVICE_ERROR,
                'Door open error',
                critical_error=True,
                device_online=True,
            ),
            '11006000': ResponseInfo(
                'BYPASS_E1_OPEN',
                ErrorTypes.DEVICE_ERROR,
                'Open circuit error',
                critical_error=True,
                device_online=True,
            ),
            '11007000': ResponseInfo(
                'BYPASS_E2_SHORT',
                ErrorTypes.DEVICE_ERROR,
                'Short circuit error',
                device_online=True,
                critical_error=True,
            ),
            '11015000': ResponseInfo(
                'BYPASS_E3_WARM',
                ErrorTypes.DEVICE_ERROR,
                'Warm error',
                critical_error=True,
                device_online=True,
            ),
            '11018000': ResponseInfo(
                'BYPASS_SET_MIST_LEVEL',
                ErrorTypes.DEVICE_ERROR,
                'Cannot set mist level error',
                device_online=True,
            ),
            '11019000': ResponseInfo(
                'BYPASS_E6_VOLTAGE_LOW',
                ErrorTypes.DEVICE_ERROR,
                'Low voltage error',
                critical_error=True,
                device_online=True,
            ),
            '11013000': ResponseInfo(
                'BYPASS_E7_VOLTAGE',
                ErrorTypes.DEVICE_ERROR,
                'Voltage error',
                critical_error=True,
                device_online=True,
            ),
            '11607000': ResponseInfo(
                'BYPASS_HUMIDIFIER_ERROR_CONNECT_MSG',
                ErrorTypes.DEVICE_ERROR,
                'Humidifier connect message error',
            ),
            '11317000': ResponseInfo(
                'BYPASS_DIMMER_NCT',
                ErrorTypes.DEVICE_ERROR,
                'Dimmer NCT error',
                critical_error=True,
                device_online=True,
            ),
            '11608000': ResponseInfo(
                'BYPASS_HUMIDIFIER_ERROR_WATER_PUMP',
                ErrorTypes.DEVICE_ERROR,
                'Humidifier water pump error',
                critical_error=True,
                device_online=True,
            ),
            '11609000': ResponseInfo(
                'BYPASS_HUMIDIFIER_ERROR_FAN_MOTOR',
                ErrorTypes.DEVICE_ERROR,
                'Humidifier fan motor error',
                critical_error=True,
                device_online=True,
            ),
            '11601000': ResponseInfo(
                'BYPASS_HUMIDIFIER_ERROR_DRY_BURNING',
                ErrorTypes.DEVICE_ERROR,
                'Dry burning error',
                critical_error=True,
                device_online=True,
            ),
            '11602000': ResponseInfo(
                'BYPASS_HUMIDIFIER_ERROR_PTC',
                ErrorTypes.DEVICE_ERROR,
                'Humidifier PTC error',
                critical_error=True,
                device_online=True,
            ),
            '11603000': ResponseInfo(
                'BYPASS_HUMIDIFIER_ERROR_WARM_HIGH',
                ErrorTypes.DEVICE_ERROR,
                'Humidifier warm high error',
                critical_error=True,
                device_online=True,
            ),
            '11604000': ResponseInfo(
                'BYPASS_HUMIDIFIER_ERROR_WATER',
                ErrorTypes.DEVICE_ERROR,
                'Humidifier water error',
                critical_error=True,
                device_online=True,
            ),
            '11907000': ResponseInfo(
                'BYPASS_LOW_WATER',
                ErrorTypes.DEVICE_ERROR,
                'Low water error',
                device_online=True,
                critical_error=True,
            ),
            '11028000': ResponseInfo(
                'BYPASS_MOTOR_OPEN',
                ErrorTypes.DEVICE_ERROR,
                'Motor open error',
                device_online=True,
                critical_error=True,
            ),
            '11017000': ResponseInfo(
                'BYPASS_NOT_SUPPORTED', ErrorTypes.REQUEST_ERROR, 'Not supported error'
            ),
            '11905000': ResponseInfo(
                'BYPASS_NO_POT',
                ErrorTypes.DEVICE_ERROR,
                'No pot error',
                device_online=True,
                critical_error=True,
            ),
            '12001000': ResponseInfo(
                'BYPASS_LACK_FOOD',
                ErrorTypes.DEVICE_ERROR,
                'Lack of food error',
                device_online=True,
                critical_error=True,
            ),
            '12002000': ResponseInfo(
                'BYPASS_JAM_FOOD',
                ErrorTypes.DEVICE_ERROR,
                'Jam food error',
                device_online=True,
                critical_error=True,
            ),
            '12003000': ResponseInfo(
                'BYPASS_BLOCK_FOOD',
                ErrorTypes.DEVICE_ERROR,
                'Block food error',
                device_online=True,
                critical_error=True,
            ),
            '12004000': ResponseInfo(
                'BYPASS_PUMP_FAIL',
                ErrorTypes.DEVICE_ERROR,
                'Pump failure error',
                device_online=True,
                critical_error=True,
            ),
            '12005000': ResponseInfo(
                'BYPASS_CALI_FAIL',
                ErrorTypes.DEVICE_ERROR,
                'Calibration failure error',
                device_online=True,
                critical_error=True,
            ),
            '11611000': ResponseInfo(
                'BYPASS_FILTER_TRAY_ERROR',
                ErrorTypes.DEVICE_ERROR,
                'Filter tray error',
                critical_error=True,
                device_online=True,
            ),
            '11610000': ResponseInfo(
                'BYPASS_VALUE_ERROR',
                ErrorTypes.DEVICE_ERROR,
                'Value error',
                critical_error=True,
                device_online=True,
            ),
            '11022000': ResponseInfo(
                'BYPASS_CANNOT_SET_LEVEL',
                ErrorTypes.DEVICE_ERROR,
                'Cannot set level error',
                critical_error=False,
                device_online=True,
            ),
            '11023000': ResponseInfo(
                'BYPASS_NTC_BOTTOM_OPEN',
                ErrorTypes.DEVICE_ERROR,
                'NTC bottom open error',
                critical_error=True,
                device_online=True,
            ),
            '11024000': ResponseInfo(
                'BYPASS_NTC_BOTTOM_SHORT',
                ErrorTypes.DEVICE_ERROR,
                'NTC bottom short error',
                critical_error=True,
                device_online=True,
            ),
            '11026000': ResponseInfo(
                'BYPASS_NTC_TOP_OPEN',
                ErrorTypes.DEVICE_ERROR,
                'NTC top open error',
                critical_error=True,
                device_online=True,
            ),
            '11025000': ResponseInfo(
                'BYPASS_NTC_TOP_SHORT',
                ErrorTypes.DEVICE_ERROR,
                'NTC top short error',
                critical_error=True,
                device_online=True,
            ),
            '11027000': ResponseInfo(
                'BYPASS_OPEN_HEAT_PIPE_OR_OPEN_FUSE',
                ErrorTypes.DEVICE_ERROR,
                'Open heat pipe or fuse error',
                critical_error=True,
                device_online=True,
            ),
            '11906000': ResponseInfo(
                'BYPASS_OVER_HEATED',
                ErrorTypes.DEVICE_ERROR,
                'Overheated error',
                critical_error=True,
                device_online=True,
            ),
            '11000000': ResponseInfo(
                'BYPASS_PARAMETER_INVALID',
                ErrorTypes.REQUEST_ERROR,
                'Invalid bypass parameter',
            ),
            '11510000': ResponseInfo(
                'BYPASS_SCHEDULE_CONFLICT', ErrorTypes.CONFIG_ERROR, 'Schedule conflict'
            ),
            '11502000': ResponseInfo(
                'BYPASS_SCHEDULE_MAX',
                ErrorTypes.CONFIG_ERROR,
                'Maximum number of schedules reached',
            ),
            '11507000': ResponseInfo(
                'BYPASS_SCHEDULE_NOT_EXIST',
                ErrorTypes.CONFIG_ERROR,
                'Schedule does not exist',
            ),
            '11503000': ResponseInfo(
                'TIMER_MAX',
                ErrorTypes.CONFIG_ERROR,
                'Maximum number of timers reached',
            ),
            '11508000': ResponseInfo(
                'TIMER_NOT_EXIST',
                ErrorTypes.CONFIG_ERROR,
                'Timer does not exist',
            ),
            '11605000': ResponseInfo(
                'BYPASS_WATER_LOCK',
                ErrorTypes.DEVICE_ERROR,
                'Water lock error',
                critical_error=True,
                device_online=True,
            ),
            '11029000': ResponseInfo(
                'BYPASS_WIFI_ERROR', ErrorTypes.DEVICE_ERROR, 'WiFi error'
            ),
            '11902000': ResponseInfo(
                'BY_PASS_ERROR_COOKING_158',
                ErrorTypes.DEVICE_ERROR,
                'Error setting cook mode, air fryer is already cooking',
                device_online=True,
            ),
            '11035000': ResponseInfo(
                'BYPASS_MOTOR_ABNORMAL_ERROR',
                ErrorTypes.DEVICE_ERROR,
                'Motor abnormal error',
                critical_error=True,
                device_online=True,
            ),
            '11903000': ResponseInfo(
                'BY_PASS_ERROR_NOT_COOK_158',
                ErrorTypes.DEVICE_ERROR,
                'Error pausing, air fryer is not cooking',
                device_online=True,
            ),
            '-12001000': ResponseInfo(
                'CONFIGKEY_EXPIRED', ErrorTypes.CONFIG_ERROR, 'Configkey expired'
            ),
            '-12000000': ResponseInfo(
                'CONFIGKEY_NOT_EXIST',
                ErrorTypes.CONFIG_ERROR,
                'Configkey does not exist',
            ),
            '-11305000': ResponseInfo(
                'CONFIG_MODULE_NOT_EXIST',
                ErrorTypes.REQUEST_ERROR,
                'Config module does not exist',
            ),
            '-11100000': ResponseInfo(
                'DATABASE_FAILED', ErrorTypes.SERVER_ERROR, 'Database error'
            ),
            '-11101000': ResponseInfo(
                'DATABASE_FAILED_ERROR', ErrorTypes.SERVER_ERROR, 'Database error'
            ),
            '-11306000': ResponseInfo(
                'DEVICE_BOUND',
                ErrorTypes.CONFIG_ERROR,
                'Device already associated with another account',
            ),
            '-11301000': ResponseInfo(
                'DEVICE_NOT_EXIST',
                ErrorTypes.CONFIG_ERROR,
                'Device does not exist',
                device_online=False,
            ),
            '-11300000': ResponseInfo(
                'DEVICE_OFFLINE',
                ErrorTypes.DEVICE_OFFLINE,
                'Device offline',
                device_online=False,
            ),
            '-11302000': ResponseInfo(
                'DEVICE_TIMEOUT',
                ErrorTypes.DEVICE_ERROR,
                'Device timeout',
                device_online=False,
            ),
            '-11304000': ResponseInfo(
                'DEVICE_TIMEZONE_DIFF',
                ErrorTypes.CONFIG_ERROR,
                'Device timezone difference',
            ),
            '-11303000': ResponseInfo(
                'FIRMWARE_LATEST',
                ErrorTypes.CONFIG_ERROR,
                'No firmware update available',
            ),
            '-11102000': ResponseInfo(
                'INTERNAL_ERROR', ErrorTypes.SERVER_ERROR, 'Internal server error'
            ),
            '-11004000': ResponseInfo(
                'METHOD_NOT_FOUND', ErrorTypes.REQUEST_ERROR, 'Method not found'
            ),
            '-11107000': ResponseInfo(
                'MONGODB_ERROR', ErrorTypes.SERVER_ERROR, 'MongoDB error'
            ),
            '-11105000': ResponseInfo(
                'MYSQL_ERROR', ErrorTypes.SERVER_ERROR, 'MySQL error'
            ),
            '88888888': ResponseInfo(
                'NETWORK_DISABLE', ErrorTypes.SERVER_ERROR, 'Network disabled'
            ),
            '77777777': ResponseInfo(
                'NETWORK_TIMEOUT', ErrorTypes.SERVER_ERROR, 'Network timeout'
            ),
            '4031005': ResponseInfo(
                'NO_PERMISSION_7A', ErrorTypes.DEVICE_ERROR, 'No 7A Permissions'
            ),
            '-11201000': ResponseInfo(
                'PASSWORD_ERROR', ErrorTypes.AUTHENTICATION, 'Invalid password'
            ),
            '-11901000': ResponseInfo(
                'PID_NOT_EXIST', ErrorTypes.DEVICE_ERROR, 'PID does not exist'
            ),
            '-11106000': ResponseInfo(
                'REDIS_ERROR', ErrorTypes.SERVER_ERROR, 'Redis error'
            ),
            '-11003000': ResponseInfo(
                'REQUEST_HIGH', ErrorTypes.RATE_LIMIT, 'Rate limiting error'
            ),
            '-11005000': ResponseInfo(
                'RESOURCE_NOT_EXIST',
                ErrorTypes.REQUEST_ERROR,
                'No device with ID found',
                device_online=False,
            ),
            '-11108000': ResponseInfo('S3_ERROR', ErrorTypes.SERVER_ERROR, 'S3 error'),
            '-11502000': ResponseInfo(
                'SCHEDULE_MAX',
                ErrorTypes.CONFIG_ERROR,
                'Maximum number of schedules reached',
            ),
            '-11103000': ResponseInfo(
                'SERVER_BUSY', ErrorTypes.SERVER_ERROR, 'Server busy'
            ),
            '-11104000': ResponseInfo(
                'SERVER_TIMEOUT', ErrorTypes.SERVER_ERROR, 'Server timeout'
            ),
            '-11501000': ResponseInfo(
                'TIMER_CONFLICT', ErrorTypes.DEVICE_ERROR, 'Timer conflict'
            ),
            '-11503000': ResponseInfo(
                'TIMER_MAX', ErrorTypes.DEVICE_ERROR, 'Maximum number of timers reached'
            ),
            '-11500000': ResponseInfo(
                'TIMER_NOT_EXIST', ErrorTypes.DEVICE_ERROR, 'Timer does not exist'
            ),
            '-11001000': ResponseInfo(
                'TOKEN_EXPIRED', ErrorTypes.TOKEN_ERROR, 'Invalid token'
            ),
            '-999999999': ResponseInfo(
                'UNKNOWN', ErrorTypes.UNKNOWN_ERROR, 'Unknown error'
            ),
            '-11307000': ResponseInfo(
                'UUID_NOT_EXIST',
                ErrorTypes.DEVICE_ERROR,
                'Device UUID not found',
                device_online=False,
            ),
            '12102000': ResponseInfo(
                'TEM_SENOR_ERROR',
                ErrorTypes.DEVICE_ERROR,
                'Temperature sensor error',
                critical_error=True,
                device_online=True,
            ),
            '12103000': ResponseInfo(
                'HUM_SENOR_ERROR',
                ErrorTypes.DEVICE_ERROR,
                'Humidity sensor error',
                critical_error=True,
                device_online=True,
            ),
            '12101000': ResponseInfo(
                'SENSOR_ERROR',
                ErrorTypes.DEVICE_ERROR,
                'Sensor error',
                critical_error=True,
                device_online=True,
            ),
            '11005000': ResponseInfo(
                'BYPASS_DEVICE_IS_OFF',
                ErrorTypes.DEVICE_ERROR,
                'Device is offDevice is off',
                critical_error=True,
                device_online=True,
            ),
        }
    )

    @classmethod
    def get_error_info(
        cls, error_code: str | int | None, msg: str | None = None
    ) -> ResponseInfo:
        """Return error dictionary for the given error code.

        Args:
            error_code (str | int): Error code to lookup.
            msg: (str | None): Optional message from API.

        Returns:
            dict: Error dictionary for the given error code.

        Example:
            ```python
            ErrorCodes.get_error_info("-11201022")
            ErrorInfo(
                "PASSWORD_ERROR", ErrorTypes.AUTHENTICATION, "Invalid password"
            )
            ```
        """
        try:
            if error_code is None:
                error_code = '-999999999'
            error_str = str(error_code)
            error_int = int(error_code)
            if error_str == '0':
                error_info = ResponseInfo('SUCCESS', ErrorTypes.SUCCESS, 'Success')
            elif error_str in cls.errors:
                error_info = cls.errors[error_str]
            else:
                error_code = int(error_int / 1000) * 1000
                error_info = cls.errors[str(error_code)]

            error_info = replace(error_info)
            if msg is not None:
                error_info.message = f'{error_info.message} - {msg}'
            error_info.code = error_int

        except (ValueError, TypeError, KeyError):
            error_info = ResponseInfo(
                'UNKNOWN', ErrorTypes.UNKNOWN_ERROR, 'Unknown error'
            )

        return error_info

    @classmethod
    def is_critical(cls, error_code: str | int) -> bool:
        """Check if error code is a device error.

        Args:
            error_code (str | int): Error code to check.

        Returns:
            bool: True if error code is a device error, False otherwise.
        """
        error_info = cls.get_error_info(error_code)
        return bool(error_info.critical_error)


class VeSyncError(Exception):
    """Base exception for VeSync errors.

    These are raised based on API response codes and exceptions
    that may be raised by the different handlers.
    """


class VeSyncLoginError(VeSyncError):
    """Exception raised for login authentication errors."""

    def __init__(self, msg: str) -> None:
        """Initialize the exception with a message."""
        super().__init__(msg)


class VeSyncTokenError(VeSyncError):
    """Exception raised for VeSync API authentication errors."""

    def __init__(self, msg: str | None = None) -> None:
        """Initialize the exception with a message."""
        if msg is None or msg.strip() == '':
            msg = 'Re-authentication required'
        super().__init__(f'Token expired or invalid - {msg}')


class VeSyncServerError(VeSyncError):
    """Exception raised for VeSync API server errors."""

    def __init__(self, msg: str) -> None:
        """Initialize the exception with a message."""
        super().__init__(msg)


class VeSyncRateLimitError(VeSyncError):
    """Exception raised for VeSync API rate limit errors."""

    def __init__(self) -> None:
        """Initialize the exception with a message."""
        super().__init__('VeSync API rate limit exceeded')


class VeSyncAPIResponseError(VeSyncError):
    """Exception raised for malformed VeSync API responses."""

    def __init__(self, msg: None | str = None) -> None:
        """Initialize the exception with a message."""
        if msg is None:
            msg = 'Unexpected VeSync API response.'
        super().__init__(msg)


class VeSyncAPIStatusCodeError(VeSyncError):
    """Exception raised for malformed VeSync API responses."""

    def __init__(self, status_code: str | None = None) -> None:
        """Initialize the exception with a message."""
        message = 'VeSync API returned an unknown status code'
        if status_code is not None:
            message = f'VeSync API returned status code {status_code}'
        super().__init__(message)


def raise_api_errors(error_info: ResponseInfo) -> None:
    """Raise the appropriate exception for API error code.

    Called by `VeSync.async_call_api` method.

    Raises:
        VeSyncRateLimitError: Rate limit error
        VeSyncLoginError: Authentication error
        VeSyncTokenError: Token error
        VeSyncServerError: Server error
    """
    match error_info.error_type:
        case ErrorTypes.RATE_LIMIT:
            raise VeSyncRateLimitError
        case ErrorTypes.AUTHENTICATION:
            raise VeSyncLoginError(error_info.message)
        case ErrorTypes.SERVER_ERROR:
            msg = (
                f'{error_info.message} - '
                'Please report error to github.com/webdjoe/pyvesync/issues'
            )
            raise VeSyncServerError(msg)



"""pytest test parametrization for VeSync devices."""
import os
import sys
import pytest


def pytest_addoption(parser):
    """Prevent new API's from being written during pipeline tests."""
    parser.addoption(
        "--write_api",
        action="store_true",
        default=False,
        help="run tests without writing API to yaml",
    )
    parser.addoption(
        "--overwrite",
        action="store_true",
        default=False,
        help="overwrite existing API in yaml - WARNING do not use unless absolutely necessary",
    )


def pytest_generate_tests(metafunc):
    """Generate tests for device methods.

    Excludes legacy tests that start with 'test_x'.
    """
    if metafunc.cls is None or 'test_x' in metafunc.module.__name__:
        return
    write_api = bool(metafunc.config.getoption('--write_api'))

    overwrite = bool(metafunc.config.getoption('--overwrite'))

    metafunc.cls.overwrite = overwrite
    metafunc.cls.write_api = write_api

    # Require class attribute 'device' to parametrize tests
    if 'device' in metafunc.cls.__dict__:
        device = metafunc.cls.__dict__['device']
        if device not in metafunc.cls.__dict__:
            return
        devices = metafunc.cls.__dict__[device]

        if metafunc.function.__name__ == 'test_details':
            return details_generator(metafunc, device, devices)

        if metafunc.function.__name__ == 'test_methods':
            return method_generator(metafunc, device, devices)


def details_generator(metafunc, gen_type, devices):
    """Parametrize device tests for get_details()."""
    id_list = []
    argvalues = []
    for setup_entry in devices:
        id_list.append(f"{gen_type}.{setup_entry}.update")
        argvalues.append([setup_entry, 'update'])
    metafunc.parametrize("setup_entry, method", argvalues, ids=id_list)


def method_generator(metafunc, gen_type, devices):
    """Parametrize device tests for methods."""
    if 'base_methods' in metafunc.cls.__dict__:
        base_methods = metafunc.cls.__dict__['base_methods']
    else:
        base_methods = []
    call_dict = {dt: base_methods.copy() for dt in devices}
    if 'device_methods' in metafunc.cls.__dict__:
        dev_methods = metafunc.cls.__dict__['device_methods']
        for dev in call_dict:
            if dev in dev_methods:
                call_dict[dev].extend(dev_methods[dev])
    id_list = []
    argvalues = []
    for dev, methods in call_dict.items():
        for method in methods:
            id_list.append(f'{gen_type}.{dev}.{method}')
            argvalues.append([dev, method])
    metafunc.parametrize("setup_entry, method", argvalues, ids=id_list)
    return


def _is_interactive(config: pytest.Config) -> bool:
    """Best-effort check for an interactive terminal."""
    tr = config.pluginmanager.getplugin("terminalreporter")
    if tr and getattr(tr, "isatty", False):
        return True
    for stream in (sys.__stdin__, sys.__stdout__, sys.__stderr__):
        try:
            if stream and stream.isatty():
                return True
        except Exception:
            pass
    return False


def _prompt_confirm(config: pytest.Config) -> bool:
    """Suspend capture, show a prompt, and return True if user confirmed."""
    flags = []
    if config.getoption("--write_api"):
        flags.append("--write_api")
    if config.getoption("--overwrite"):
        flags.append("--overwrite")

    if len(flags) == 0:
        return True

    if '--overwrite' in flags and '--write_api' not in flags:
        # Overwrite requires write_api to be set
        pytest.exit(
            "\nBoth --overwrite and --write_api need to be set to overwrite API calls.", 1
        )

    # Interactive prompt
    if "--write_api" in flags and "--overwrite" not in flags:
        msg = (
            "\nAre you sure you want to write API requests to files?\n"
            "Continue? [y/N]: "
        )
    elif "--overwrite" in flags and "--write_api" in flags:
        msg = (
            "\nAre you sure you want to OVERWRITE existing API requests?\n"
            "Continue? [y/N]: "
        )
    else:
        pytest.exit("\nUnrecognized combination of flags: " + ", ".join(flags), 1)

    if not flags:
        return True
    capman = config.pluginmanager.getplugin("capturemanager")
    tr = config.pluginmanager.getplugin("terminalreporter")
    if capman:
        # allow stdin/stdout to pass through for input()
        capman.suspend_global_capture(in_=True)  # type: ignore[arg-type]
    try:
        if tr and getattr(tr, "isatty", False):
            tr._tw.write(msg)   # type: ignore[arg-type]
        else:
            print(msg, end="", flush=True)
        try:
            resp = input().strip().lower()
        except EOFError:
            resp = ""
        return resp in {"y", "yes"}
    finally:
        if capman:
            capman.resume_global_capture()  # type: ignore[arg-type]


def _require_confirmation(config: pytest.Config) -> None:
    """Prompt once per session if any destructive flags are set."""
    if not config.getoption("--write_api") and not config.getoption("--overwrite"):
        return

    # Allow opt-out via flag or env var in CI
    if str(os.environ.get("PYTEST_CONFIRM", "")).lower() in {"1", "true", "yes", "y"}:
        return

    # If there's no TTY, force explicit opt-out instead of hanging
    if not _is_interactive(config):
        raise pytest.UsageError(
            "write_api require confirmation, but the session is non-interactive. "
            "Re-run with environment variable set PYTEST_CONFIRM=1."
        )

    if not _prompt_confirm(config):
        pytest.exit("Aborted by user.", returncode=1)


def pytest_sessionstart(session: pytest.Session) -> None:
    # Runs before collection; safe place to ask once and abort if needed
    _require_confirmation(session.config)



import orjson
from pyvesync.devices.vesyncoutlet import VeSyncOutlet15A
from pyvesync.utils.helpers import Helpers as Helpers
import call_json
import base_test_cases
import call_json_outlets

DEVICE_TYPE = 'ESW15-USA'

DEV_LIST_DETAIL = call_json.DeviceList.device_list_item(DEVICE_TYPE)

CORRECT_15A_LIST = call_json.DeviceList.device_list_response(DEVICE_TYPE)

ENERGY_HISTORY = call_json_outlets.ENERGY_HISTORY

CORRECT_15A_DETAILS = call_json_outlets.DETAILS_RESPONSES[DEVICE_TYPE]

BAD_15A_LIST = call_json.DETAILS_BADCODE


class TestVeSyncSwitch(base_test_cases.TestBase):

    def test_15aswitch_conf(self):
        """Tests that 15A Outlet is instantiated properly"""
        self.mock_api.return_value = CORRECT_15A_LIST
        self.run_in_loop(self.manager.get_devices)
        outlets = self.manager.outlets
        assert len(outlets) == 1
        vswitch15a = outlets[0]
        assert isinstance(vswitch15a, VeSyncOutlet15A)
        assert vswitch15a.device_name == call_json.TestDefaults.name(DEVICE_TYPE)
        assert vswitch15a.device_type == DEVICE_TYPE
        assert vswitch15a.cid == call_json.TestDefaults.cid(DEVICE_TYPE)
        assert vswitch15a.uuid == call_json.TestDefaults.uuid(DEVICE_TYPE)

    def test_15a_details(self):
        """Test 15A get_details() """
        resp_dict, status_code = CORRECT_15A_DETAILS
        self.mock_api.return_value = orjson.dumps(resp_dict), status_code
        vswitch15a = VeSyncOutlet15A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch15a.get_details)
        dev_details = vswitch15a.details
        assert vswitch15a.device_status == 'on'
        assert isinstance(dev_details, dict)
        assert dev_details['active_time'] == 1
        assert dev_details['energy'] == 1
        assert dev_details['power'] == '1'
        assert dev_details['voltage'] == '1'
        assert vswitch15a.power == 1
        assert vswitch15a.voltage == 1
        assert vswitch15a.active_time == 1
        assert vswitch15a.energy_today == 1


class TestVesync15ASwitch(base_test_cases.TestBase):

    def test_15aswitch_conf(self):
        """Tests that 15A Outlet is instantiated properly"""
        self.mock_api.return_value = CORRECT_15A_LIST
        self.run_in_loop(self.manager.get_devices)
        outlets = self.manager.outlets
        assert len(outlets) == 1
        vswitch15a = outlets[0]
        assert isinstance(vswitch15a, VeSyncOutlet15A)
        assert vswitch15a.device_name == call_json.TestDefaults.name(DEVICE_TYPE)
        assert vswitch15a.device_type == DEVICE_TYPE
        assert vswitch15a.cid == call_json.TestDefaults.cid(DEVICE_TYPE)
        assert vswitch15a.uuid == call_json.TestDefaults.uuid(DEVICE_TYPE)

    def test_15a_details_fail(self):
        """Test 15A get_details with Code>0"""
        resp_dict, status_code = BAD_15A_LIST
        self.mock_api.return_value = orjson.dumps(resp_dict), status_code
        vswitch15a = VeSyncOutlet15A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch15a.get_details)
        assert len(self.caplog.records) == 1
        assert 'details' in self.caplog.text

    def NO_test_15a_no_details(self):
        """Test 15A details return with no details and code=0"""
        # Removed test - will not be needed with API response validation
        bad_15a_details = {'code': 0, 'deviceStatus': 'on'}
        self.mock_api.return_value = (orjson.dumps(bad_15a_details), 200)
        vswitch15a = VeSyncOutlet15A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch15a.get_details)
        assert len(self.caplog.records) == 1

    def test_15a_onoff(self):
        """Test 15A Device On/Off Methods"""
        self.mock_api.return_value = (orjson.dumps({'code': 0}), 200)
        vswitch15a = VeSyncOutlet15A(DEV_LIST_DETAIL, self.manager)
        head = Helpers.req_legacy_headers(self.manager)
        body = Helpers.req_body(self.manager, 'devicestatus')

        body['status'] = 'on'
        body['uuid'] = vswitch15a.uuid
        on = self.run_in_loop(vswitch15a.turn_on)
        self.mock_api.assert_called_with(
            '/15a/v1/device/devicestatus', 'put', headers=head, json_object=body
        )
        assert on
        off = self.run_in_loop(vswitch15a.turn_off)
        body['status'] = 'off'
        self.mock_api.assert_called_with(
            '/15a/v1/device/devicestatus', 'put', headers=head, json_object=body
        )
        assert off

    def test_15a_onoff_fail(self):
        """Test 15A On/Off Fail with Code>0"""
        self.mock_api.return_value = (orjson.dumps({'code': 1}), 400)
        vswitch15a = VeSyncOutlet15A(DEV_LIST_DETAIL, self.manager)
        assert not self.run_in_loop(vswitch15a.turn_on)
        assert not self.run_in_loop(vswitch15a.turn_off)

    def test_15a_weekly(self):
        """Test 15A get_weekly_energy"""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        vswitch15a = VeSyncOutlet15A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch15a.get_weekly_energy)
        body = Helpers.req_body(self.manager, 'energy_week')
        body['uuid'] = vswitch15a.uuid
        self.mock_api.assert_called_with(
            '/15a/v1/device/energyweek',
            'post',
            headers=Helpers.req_legacy_headers(self.manager),
            json_object=body,
        )
        energy_dict = vswitch15a.energy['week']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]
        assert vswitch15a.weekly_energy_total == 1

    def test_15a_monthly(self):
        """Test 15A get_monthly_energy"""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        vswitch15a = VeSyncOutlet15A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch15a.get_monthly_energy)
        body = Helpers.req_body(self.manager, 'energy_month')
        body['uuid'] = vswitch15a.uuid
        self.mock_api.assert_called_with(
            '/15a/v1/device/energymonth',
            'post',
            headers=Helpers.req_legacy_headers(self.manager),
            json_object=body,
        )
        energy_dict = vswitch15a.energy['month']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]
        assert vswitch15a.monthly_energy_total == 1

    def test_15a_yearly(self):
        """Test 15A get_yearly_energy"""
        resp_dict, status = ENERGY_HISTORY
        self.mock_api.return_value = orjson.dumps(resp_dict), status
        vswitch15a = VeSyncOutlet15A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch15a.get_yearly_energy)
        body = Helpers.req_body(self.manager, 'energy_year')
        body['uuid'] = vswitch15a.uuid
        self.mock_api.assert_called_with(
            '/15a/v1/device/energyyear',
            'post',
            headers=Helpers.req_legacy_headers(self.manager),
            json_object=body,
        )
        energy_dict = vswitch15a.energy['year']
        assert energy_dict['energy_consumption_of_today'] == 1
        assert energy_dict['cost_per_kwh'] == 1
        assert energy_dict['max_energy'] == 1
        assert energy_dict['total_energy'] == 1
        assert energy_dict['data'] == [1, 1]
        assert vswitch15a.yearly_energy_total == 1

    def test_history_fail(self):
        """Test 15A energy failure"""
        bad_history = {'code': 1}
        self.mock_api.return_value = (orjson.dumps(bad_history), 200)
        vswitch15a = VeSyncOutlet15A(DEV_LIST_DETAIL, self.manager)
        self.run_in_loop(vswitch15a.update_energy)
        assert len(self.caplog.records) == 1
        assert 'weekly' in self.caplog.text
        self.caplog.clear()
        self.run_in_loop(vswitch15a.get_monthly_energy)
        assert len(self.caplog.records) == 1
        assert 'monthly' in self.caplog.text
        self.caplog.clear()
        self.run_in_loop(vswitch15a.get_yearly_energy)
        assert len(self.caplog.records) == 1
        assert 'yearly' in self.caplog.text
